(define export '(export
                 (forms doc-cash-book
                        arch-cash-book)))

(define ds-cash-book
  (quasiquote
   (dataset
    (extend
     (Note
      (mode extend)
      (index 0)
      (type ExtString)
      (caption "Примечание")
      (tags template))
     (From
      (mode extend)
      (index 1)
      (type Date)
      (caption "Период с"))
     (CurrentDate
      (mode extend)
      (index 2)
      (type Date)
      (caption "Дата листа"))
     (Rest
      (mode extend)
      (index 3)
      (type Money)
      (caption "Остаток на начало периода"))
     (PrintMethod
      (mode extend)
      (index 4)
      (type ShortString)
      (caption "к печати")))
    (fields (Doc SumMethod Precision Type Curr ByCurr Firm FirmCargo Partner PartnerCargo Date Name Num Sum
                 Note From CurrentDate Rest PrintMethod)))))

(define ds-doc-cb
  (quasiquote
   (dataset
    (extend
     (CreditSum
      (mode raw)
      (type Money)
      (raw "case when r_Doc.\"fType\" = 'doc-credit-order' then r_Doc.\"fSum\" else 0.0 end"))
     (AccSum
      (mode raw)
      (type Money)
      (raw "case when r_Doc.\"fType\" != 'doc-credit-order' then r_Doc.\"fSum\" else 0.0 end")))
    (fields (Doc Num Name Date Type Firm (Partner ShortName) Sum
                 CreditSum AccSum))
    (where (Firm = :aFirm)
           (Date = :aCurrentDate)
           (Date > "'1899-12-31'")
           (Type IN "('doc-credit-order', 'doc-acc-order')")
           ))))

(define-doc doc-cash-book
  (quasiquote
   (form
    (caption "Кассовая книга")
    (dataset ds-cash-book)
    (archive arch-cash-book)
    (controls
     (Goods
      (type Grid)
      (dataset ds-doc-cb)
      (bind ([aFirm Master Firm]
             [aCurrentDate Master CurrentDate]
             ))
      
      (columns ([Name "Номер"]
                [Date "Дата"]
                ;[Type "Тип"]
                [PartnerShortName "Партнер" 200]
                [CreditSum "Приход"]
                [AccSum "Расход"]))
      )
     ,mutual-firm)
    (visual
     (Date
      (caption "по"))
     ,(ctrl-firm 'Firm $doc 'duties-std "Организация")
     (PrintMethod
      (type Combo)
      (mode string)
      (values (["<empty>" "Вся кассовая книга"]
               ["Current" "На выбранную дату"]
               ["Title" "Титульный лист"]
               ["Total" "Последняя страница"]))))
    (layout (v (h Firm ">Примечание" Name)
               (h From Date Rest)
               (h CurrentDate PrintMethod)
               Goods))
    (groups ([Firm From CurrentDate]))
    (styles ([Mini Precision Rest])))))

(define arch-cash-book
  (quasiquote
   (form
    (caption "Архив кассовых книг")
    (dataset ds-arch)
    (extend ArchCommon)
    (controls
     (Docs
      (type Grid)
      (dataset (ds-arch-docs 'doc-cash-book))
      (bind ([aFirm Master Firm]
             [aPartner Master Partner]
             [aFrom Master From]
             [aTo Master To]
             [aPayState Master PayState]))
      (form null)
      (columns ([Mark "Метка" 25 rw]
                [Name "Примечание"]
                [Date "Конец периода" 80]
                [FirmShortName "Организация" 200]
                ;[Sum "Сумма"]
                )))
     (CancelRange
      (type NullButton)
      (bind ([Firm Master Firm]
             [Partner Master Partner]
             [From Master From]
             [To Master To]
             [PayState Master PayState]))))
    (visual
     (Firm
      (dataset ds-firm-min)
      (bind (auto-name
             [aSelf Const True]))
      (form null))
     (Partner
      (dataset ds-partner-min)
      (form null)))
    (layout (t ("Документы")
               (v (h CancelRange
                     (v From To)
                     (h "Организация" Firm))
                  Docs)))
    (groups ([From To] ["Организация" ">Партнер"])))))
