(define export '(export
                 (forms form-volume-dict
                        form-group-dict
                        form-model-dict
                        form-firm-dict
                        form-country-dict
                        form-curr-dict
                        ;                        form-dutie-dict
                        ;                        form-people-dict
                        )))

(define form-volume-dict
  (quasiquote
   (form
    (caption "Единицы измерения")
    (archive dictionary)
    (dataset ds-dict)
    (controls
     (Volume
      (type Grid)
      (dataset ds-volume-min)
      (columns ([Name "Наименование" 450]))
      (form form-volume-std)
      (combine enabled)))
    (layout (v Volume)))))

(define form-group-dict
  (quasiquote
   (form
    (caption "Группы товаров")
    (archive dictionary)
    (dataset ds-dict)
    (controls
     (Groups
      (type Grid)
      (dataset ds-group-min)
      (columns ([Name "Наименование" 450]))
      (form form-group-std)
      (combine enabled)))
    (layout (v Groups)))))

(define ds-model-dict
  (quasiquote
   (dataset
    (fields (Model Name Price (Group Name) (Volume Name))))))

(define ds-model-dict-by-group
  (quasiquote
   (dataset
    (fields (Model Name Price (Group Name) (Volume Name)))
    (where (Object > "'000000000000'")
           [nnvl Group = :aGroup]))))

(define form-model-dict
  (quasiquote
   (form
    (caption "Товары")
    (archive dictionary)
    (dataset ds-dict)
    (controls
     (Group
      (type Reference)
      (dataset ds-group-min)
      (form form-group-std)
      (caption "Группа"))
     (Models
      (type Grid)
      (dataset ds-model-dict-by-group)
      (bind ([aGroup Group Default]))
      (columns ([Name "Наименование" 450]
                [VolumeName "Ед. изм."]
                [Price "Цена"]))
      (form form-model-std)
      (combine enabled)))
    (layout (v Group Models)))))

(define form-firm-dict
  (quasiquote
   (form
    (caption "Фирмы")
    (archive dictionary)
    (dataset ds-dict)
    (controls
     (Firms
      (type Grid)
      (dataset ds-firm-std)
      (columns ([Self "Пл" 20]
                [ShortName "Фирма" 150]
                [FullName "Полное название" 450]))
      (form form-firm-std-dict)
      (combine enabled)))
    (layout (v Firms)))))

(define form-firm-std-dict
  (quasiquote
   (form
    (caption "Фирма")
    (dataset ds-firm-std)
    (controls
     (ExtraFill
      (type ExtraFill)
      (bind ([FullName Master FullName]
             [ExtraName Master ExtraName])))
     (NDSFill
      (type NDSFill)
      (bind ([FullName Master FullName]
             [Address Master Address]
             [INN Master INN]
             [KPP Master KPP]
             [IpOGRN Master IpOGRN])))
     (Cargo
      (type Reference)
      (dataset ds-cargo-min)
      (bind ([aFirm Master Object]))
      (form form-cargo-std)
      (group v0))
     (SelectBank
      (type SelectBank)
      (bind ([Bank Master Bank]
             [BIK Master BIK]
             [CorDep Master CorDep]))))
    (visual
     (ExtraName
      (actions ([ExtraFill "Заполнить автоматически"])))
     (INN
      (actions ([NDSFill "Заполнить реквизиты по ИНН"]))))
    (layout (v ShortName FullName ExtraName (h ">Грузоотправитель" Cargo) Self Address INN KPP RasDep CorDep Bank BIK OKPO Custom1 Custom2 (h No2ndSign NoSeal) NoNDS))
    (childs
     (ofk "ОФК" (v OfkOrg OfkCodeOrg OfkDep OfkINN OfkCode OfkOSB OfkNBudg OfkNameFO OfkGlavaGRS OfkNameUBPGRS OfkBudgLevel OfkCodeUBP OfkNameUBP))
     (ip "ИП" (v IpOGRN (h "Свидетельство" "<>--") IpSvSer IpSvNum IpSvGive IpSvDate (h "Паспорт" "<>--") IpPspSer IpPspNum IpPspGive IpPspDate)))
    (groups ([ShortName FullName ExtraName ">Грузоотправитель" Self INN KPP RasDep CorDep Bank BIK OKPO NoNDS No2ndSign Address Custom1 Custom2])))))

(define form-country-dict
  (quasiquote
   (form
    (caption "Страны производители")
    (archive dictionary)
    (dataset ds-dict)
    (controls
     (Country
      (type Grid)
      (dataset ds-country-min)
      (columns ([Name "Наименование" 450]))
      (form form-country-std)
      (combine enabled)))
    (layout (v Country)))))

(define form-curr-dict
  (quasiquote
   (form
    (caption "Дополнительные валюты")
    (archive dictionary)
    (dataset ds-dict)
    (controls
     (Curr
      (type Grid)
      (dataset ds-curr-std)
      (columns ([Name "Обозначение валюты" 450]
                ))
      (form form-curr-std)))
    (layout (v Curr)))))

(define form-dutie-dict
  (quasiquote
   (form
    (caption "Должности")
    (archive dictionary)
    (dataset ds-dict)
    (controls
     (Dutie
      (type Grid)
      (dataset ds-dutie-min)
      (columns ([Name "Наименование" 450]))
      (form form-dutie-std)))
    (layout (v Dutie)))))

(define form-people-dict
  (quasiquote
   (form
    (caption "Представители")
    (archive dictionary)
    (dataset ds-dict)
    (controls
     (People
      (type Grid)
      (dataset ds-people-dict)
      (bind ([aFirm Master Firm]))
      (columns ([SecondName "Фамилия" 200]
                [FirstName "Имя" 200]
                [LastName "Отчество" 200]))
      (form form-people-std)))
    (visual
     (Firm
      (type Reference)
      (dataset ds-firm-std)
      ;(bind ([Default Master Firm]))
      (form null)))
    (layout (v (h "Фирма" Firm) People)))))