(define export '(export
                 (forms doc-pay-order
                        arch-pay-order)))

(define ds-pay-order
  (quasiquote
   (dataset
    (extend
     (PayType
      (mode extend)
      (index 0)
      (type ShortString)
      (caption "Вид платежа"))
     (State
      (mode extend)
      (index 1)
      (type ShortString)
      (caption "Статус"))
     (PayKind
      (mode extend)
      (index 2)
      (type ShortString))
     (PayDate
      (mode extend)
      (index 3)
      (type Date))
     (PayOrd
      (mode extend)
      (index 4)
      (type ShortString))
     (PayCode
      (mode extend)
      (index 5)
      (type ShortString))
     (PClass
      (mode extend)
      (index 6)
      (type KBK))
     (POKAT
      (mode extend)
      (index 7)
      (type Okato))
     (PBase
      (mode extend)
      (index 8)
      (type ShortString))
     (PPeriod
      (mode extend)
      (index 9)
      (type ShortString))
     (PNum
      (mode extend)
      (index 10)
      (type ShortString))
     (PDocDt
      (mode extend)
      (index 11)
      (type ShortString))
     (PKind
      (mode extend)
      (index 12)
      (type ShortString))
     (Purpose
      (mode extend)
      (index 13)
      (type MemoString))
     (NDS
      (mode extend)
      (index 14)
      (type Percent)
      (caption "НДС"))
     (No2ndSign
      (mode extend)
      (index 15)
      (type LongString))
     (NoNDS
      (mode extend)
      (index 16)
      (type LongString))
     (NDSStr
      (mode extend)
      (index 17)
      (type LongString))
     )
    (fields (Doc SumMethod Precision Type Curr ByCurr Firm FirmCargo Partner PartnerCargo Date Name Num Sum
                 PayKind PayType State PayDate PayOrd PayCode
                 PClass POKAT PBase PPeriod PNum PDocDt PKind
                 Purpose NDS No2ndSign NoNDS NDSStr)))))

(define-doc doc-pay-order
  (quasiquote
   (form
    (caption "Платежное поручение")
    (dataset ds-pay-order)
    (archive arch-pay-order)
    (fill
     (caption "платежному поручению")
     (require Nonexistent)
     (provide Payment))
    (controls
     ;(CurrCtl
     ; (type Curr)
     ; (bind ([Doc Master Object]
     ;        [Curr Master Curr]
     ;        [ByCurr Master ByCurr])))
     (Numerator
      (type Numerator)
      (bind ([Name Master Name]
             [Num Master Num])))
     (NSR
      (type Relay)
      (method "nds-string-po")
      (bind ([Sum Master Sum]
             [NDS Master NDS]
             [Str Master NDSStr]
             [NoNDS Master NoNDS])))
     (ClassCodes
      (type CmdButton)
      (text "Коды классификации")
      (cmd BuiltIn BudClass))
     (PurposeTemplate
      (type CmdButton)
      (text "Шаблон")
      (fill horizontal)
      (cmd Template)
      (template-name Doc.Purpose)
      (bind ([Template Master Purpose])))
     ,mutual-firm
     ,mutual-partner)
    (visual
     (Name
      (numerator 1))
     (Date
      (caption "от"))
     ,(ctrl-firm 'Firm $doc 'duties-std "<Плательщик")
     ,(ctrl-firm 'Partner $doc 'duties-std "<Получатель")
     (SumMethod
      (type Combo)
      (mode string)
      (values (["AA" "От суммы"]
               ["AB" "От единицы"]
               ["BA" "От суммы (банковское округление)"]
               ["BB" "От единицы (банковское округление)"]
               [<custom.calc-method> "Дополнительно..."])))
     (PayType
      (values (["Почтой"]
               ["Телеграфом"]
               ["Электронно"]
               ["Срочный"]
               ["По договору"]
               ["Клиринг"])))
     (PClass
      (dataset ds-kbk-inf)
      (bind (auto-name [aTypeKBK Const "pay-order"]))
      (form form-kbk-std))
     (POKAT
      (dataset ds-okato-min)
      (form form-okato-std))
     (PBase
      (values (["ТП"]
               ["ЗД"]
               ["ТР"]
               ["РС"]
               ["ОТ"]
               ["РТ"]
               ["ВУ"]
               ["ПР"]
               ["АП"]
               ["АР"])))
     (PKind
      (values (["НС"]
               ["ПЛ"]
               ["ГП"]
               ["ВЗ"]
               ["АВ"]
               ["ПЕ"]
               ["ПЦ"]
               ["СА"]
               ["АШ"]
               ["ИШ"]))
      (fill horizontal)))
    (layout (v (h "Платежное поручение №" Name Date PayType State)
               (h "Сумма\nпрописью" "|" "[Master.Sum/words/[rw] [rc] [kv] [kc]]")
               "--"
               (h (v (h (v Firm
                           "ИНН [Master.Firm.INN]/КПП [Master.Firm.KPP]"
                           "[Master.Firm.FullName]"
                           "--"
                           "Банк плательщика" "[Master.Firm.Bank]"
                           "--")
                        (v (h "|" "<<210>>")
                           (h "|" "--")
                           (h "|" "Сч. №" "[Master.Firm.RasDep]") "--"
                           (h "|" "БИК" "[Master.Firm.BIK]")
                           (h "|" "Сч. №" "[Master.Firm.CorDep]") "--"))
                     (h (v Partner
                           "ИНН [Master.Partner.INN]/КПП [Master.Partner.KPP]"
                           "[Master.Partner.FullName]"
                           "--"
                           "Банк получателя" "[Master.Partner.Bank]"
                           "--")
                        (v "|" (h "|" "--")
                           (h "|" "Сч. №" "[Master.Partner.RasDep]") "--"
                           (h "|" "БИК" "[Master.Partner.BIK]")
                           (h "|" "Сч. №" "[Master.Partner.CorDep]") "--")))
                  "|"
                  (v "Сумма"
                     Sum
                     "Вид оп."
                     PayKind
                     "Срок плат."
                     PayDate
                     "Очер. плат."
                     PayOrd
                     "Код"
                     PayCode
                     ))
               (h (v "КБК" PClass)
                  (v "ОКАТО" POKAT)
                  (v "Осн." PBase)
                  (v "Период" PPeriod)
                  (v "Номер" PNum)
                  (v "Дата" PDocDt)
                  (v "Тип" PKind))
               "Назначение платежа"
               Purpose
               (h NDS "%" NDSStr ClassCodes PurposeTemplate)
               ;(h CurrCtl)
               ))
    (childs
     (setup "Настройка" (v "Формат указателя отсутствия второй подписи:" No2ndSign "Формат указателя отсутствия НДС:" NoNDS))
     (class-code "Коды бюджетной классификации" (v "Формат указателя отсутствия второй подписи:" No2ndSign "Формат указателя отсутствия НДС:" NoNDS)))
    (styles ([Mini Name PayType Sum PayKind PayOrd PayCode POKAT NDS]
             [Micro State]
             [Combo2Ch PBase PKind]
             [Date10Ch PPeriod PDocDt]
             [HalfMini PNum]))
    (groups ([Firm Partner] ["БИК" "Сч. №"] [NDS]))
    )))

(define arch-pay-order (arch-form "Архив платежных поручений" 'doc-pay-order #f #f '(export epd)))
