(define export '(export
                 (forms doc-price-list
                        arch-price-list)))

(define ds-price-list
  (quasiquote
   (dataset
    (extend
     (Note
      (mode extend)
      (index 0)
      (type ExtString)
      (caption "Примечание")
      (tags template)))
    (fields (Doc SumMethod Precision Type Curr ByCurr Firm FirmCargo Partner PartnerCargo Date Name Num Note Sum)))))

(define ds-goods-price-list
  (quasiquote
   (dataset
    (fields (Goods OrderA Doc Count (Volume Name) Price NDS NDSMode (Model Name Curr (Volume Name) (Group Name)) Country GTD QName))
    (ignore-ref Doc)
    (transform GoodsCost GoodsName)
    (where (Doc = :aDoc)
           [nnvl ModelGroup = :aAutoFilter_ModelGroup]
           ))))

(define form-goods-price-list
  (quasiquote
   (form
    (tag ab bc cd)
    (caption "Товар")
    (dataset ds-form-goods)
    (controls
     (ModelNotifier
      (type ObjectNotifier)
      (bind ([Object Master Model]
             [Volume Master Volume]
             [Country Master Country]
             [GTD Master GTD]
             [Price Master Price]
             [NDS Master NDS]
             [NDSMode Master NDSMode]))))
    (visual
     (NDSMode
      (type Switch)
      (caption none)
      (values ([0 "нет"] [1 "начислить"] [-1 "выделить"])))
     )
    (layout (v Model Country GTD Volume "--" Count Price (h NDS NDSMode)))
    (groups ([Model Country GTD Volume Count Price NDS]))
    (styles ([Mini Price Count NDS]))
    )))

(define ds-model-price-list
  (quasiquote
   (dataset
    (extend
     (GoodsRef
      (mode raw)
      (type Reference)
      (raw "(SELECT FIRST 1 \"fObject\" FROM \"tGoods\" g WHERE g.\"fDoc\" = :aDoc AND g.\"fModel\" = r_Model.\"fObject\")")
      )
     (Inserted
      (mode raw)
      (type Boolean)
      (raw "(CASE WHEN EXISTS (SELECT \"fObject\" FROM \"tGoods\" g WHERE g.\"fDoc\" = :aDoc AND g.\"fModel\" = r_Model.\"fObject\") THEN '1' ELSE '0' END)")
      ))
    (fields (Model Name Price NDS NDSMode (Group Name) (Volume Name) Country GTD GoodsRef Inserted))
    (where [Object != "'000000000000'"]
           [nnvl Group = :aAutoFilter_Group]
           [nnvl Inserted = :aAutoFilter_Inserted]
           )
    )))

(define (DocPriceList.copy-fields) '(Volume Country GTD Price NDS NDSMode))

(define-doc doc-price-list
  (quasiquote
   (form
    (caption "Прайс-лист")
    (dataset ds-price-list)
    (extend DocPriceList)
    (archive arch-price-list)
    (controls
     (Goods
      (type Grid)
      (dataset ds-goods-price-list)
      (bind ([aDoc Master Object]
             [Precision Master Precision]
             [SumMethod Master SumMethod]
             [Curr CurrCtl Curr]
             [ByCurr CurrCtl ByCurr]
             [Rate CurrCtl Rate]
             [Notify.Firm Master Firm]))
      (form form-goods-price-list)
      (columns ([MModelName "Наименование" 200 rw]
                [VolumeName "Ед. изм." 100]
                [Count "Количество" 80 rw]
                [PriceF "Цена без НДС" -85]
                [PriceT "Цена"]
                [ModelGroupName "Группа"])))
     (Models
      (type Grid)
      (dataset ds-model-price-list)
      (bind ([aDoc Master Object]))
      (columns ([Inserted " " 20]
                [Name "Наименование" 450]
                [VolumeName "Ед. изм." 100]
                [Price "Цена"]
                [GroupName "Группа"]))
      (form form-model-std))
     (CurrCtl
      (type Curr)
      (bind ([Doc Master Object]
             [Curr Master Curr]
             [ByCurr Master ByCurr])))
     (Numerator
      (type Numerator)
      (bind ([Name Master Name]
             [Num Master Num])))
     ,mutual-firm)
    (visual
     (Name
      (numerator 1)
      (caption "Название"))
     (Date
      (caption "от"))
     (Note
      (type String)
      (caption "Примечание"))
     ,(ctrl-firm 'Firm $doc 'duties-std "Фирма")
     (SumMethod
      (type Combo)
      (mode string)
      (values (["AA" "От суммы"]
               ["AB" "От единицы"]
               ["BA" "От суммы (банковское округление)"]
               ["BB" "От единицы (банковское округление)"]
               [<custom.calc-method> "Дополнительно..."]))))
    (layout (v (h Firm)
               (h Name Date Note CurrCtl)
               Goods
               Models))
    (childs
     (setup "Настройка" (v Precision SumMethod)))
    (groups ([Firm Name]))
    (styles ([Mini Name Precision])))))

(define arch-price-list (arch-form "Архив прайс-листов" 'doc-price-list #t #f '(hide-partner #t)))
