(define (ds-table table where)
  (let ([fields (map car (al.val meta 'tables table 'fields))])
    (if where
        (quasiquote
         (dataset
          (fields ,(cons table fields))
          ,(cons 'where where)))
        (quasiquote
         (dataset
          (fields ,(cons table fields)))))))

(define ds-model-std
  (quasiquote
   (dataset
    (fields (Model Name Price Volume Curr NDS NDSMode GTD Country Group Note1 Note2 Pack Coefficient)))))

(define ds-goods-std
  (quasiquote
   (dataset
    (fields (Goods OrderA Doc Count Price PriceDC NDS NDSMode (Model Name Curr (Volume Name) (Group Name)) (Volume Name) QName))
    (accept-paste ds-goods-std)
    (aggregates
     (AggSum1 Sum Sum)
     (AggSumN SumN Sum))
    (ignore-ref Doc)
    (transform GoodsCost GoodsName)
    (where (Doc = :aDoc)
           [nnvl ModelGroup = :aAutoFilter_ModelGroup]
           ))))

(define ds-form-goods
  (quasiquote
   (dataset
    (ignore-ref Doc)
    (transform GoodsCost GoodsName)
    (fields (Goods Doc Count Price PriceDC NDS NDSMode (Model Name Curr Group) QName Volume GTD Country Pack Coefficient)))))

(define ds-firm-std
  (quasiquote
   (dataset
    (fields (Firm ShortName FullName INN KPP RasDep CorDep Bank BIK Self Chief Accountant Cashier OKPO NoSeal No2ndSign Address
                  OfkOrg OfkDep OfkINN OfkCode OfkOSB OfkNBudg OfkGlavaGRS OfkNameUBPGRS OfkNameFO OfkCodeOrg
                  OfkBudgLevel OfkCodeUBP OfkNameUBP Custom1 Custom2 NoNDS ExtraName
                  IpOGRN IpSvSer IpSvNum IpSvGive IpSvDate IpPspSer IpPspNum IpPspGive IpPspDate)))))

(define ds-firm-doc
  (quasiquote
   (dataset
    (fields (Firm ShortName FullName INN KPP RasDep CorDep Bank BIK Self OKPO OKDP NoSeal No2ndSign NoNDS Address Custom1 Custom2 
                  OfkOrg OfkDep OfkINN OfkCode OfkOSB OfkNBudg OfkGlavaGRS OfkNameUBPGRS OfkNameFO OfkCodeOrg
                  OfkBudgLevel OfkCodeUBP OfkNameUBP 
                  IpOGRN IpSvSer IpSvNum IpSvGive IpSvDate IpPspSer IpPspNum IpPspGive IpPspDate
                  ExtraName)))))

(define ds-people-min
  (quasiquote
   (dataset
    (extend
     (FIO
      (mode raw)
      (type LongString)
      (raw "SUBSTRING(r_People.\"fSecondName\" FROM 1 FOR 240) || CASE WHEN strlen(r_People.\"fFirstName\") > 0 THEN ' ' || SUBSTRING(r_People.\"fFirstName\" FROM 1 FOR 1) || '.' ELSE '' END || CASE WHEN STRLEN(r_People.\"fLastName\") > 0 THEN ' ' || SUBSTRING(r_People.\"fLastName\" FROM 1 FOR 1) || '.' ELSE '' END")))
    (fields (People FIO SecondName Firm))
    (where (Firm = :aFirm)))))

(define ds-people-std
  (quasiquote
   (dataset
    (fields (People FirstName SecondName LastName Firm Dutie Address TabNum
                    PspSer PspGive PspDate PspNum PspType
                    CertSer CertNum CertClass CertGive CertDate)))))

(define ds-people-form
  (quasiquote
   (dataset
    (fields (People FirstName SecondName LastName Firm Address TabNum 
                    PspSer PspGive PspDate PspNum PspType
                    CertSer CertNum CertClass CertGive CertDate Car
                    Dutie)))))

(define ds-people-dict
  (quasiquote
   (dataset
    (fields (People FirstName SecondName LastName Firm Dutie Address PspSer PspGive PspDate PspNum))
    (where (Firm = :aFirm))
    )))

(define ds-cargo-min
  (quasiquote
   (dataset
    (fields (Cargo Name Firm))
    (where (Firm = :aFirm)))))

(define ds-cargo-std
  (quasiquote
   (dataset
    (fields (Cargo Name Firm INN KPP BIK Bank CorDep RasDep OKPO Address)))))

(define ds-volume-min
  (quasiquote
   (dataset
    (fields (Volume Name OKEI)))))

(define ds-dutie-min
  (quasiquote
   (dataset
    (fields (Dutie Name)))))

(define ds-country-min
  (quasiquote
   (dataset
    (fields (Country Name)))))

(define ds-group-min
  (quasiquote
   (dataset
    (fields (Group Name)))))

(define ds-template
  (quasiquote
   (dataset
    (fields (Template Type Content))
    (where (Type = :aType)))))

(define ds-curr-std
  (quasiquote
   (dataset
    (fields (Curr Name Rate Lex_i_a Lex_i_b Lex_i_c Lex_i_s Lex_f_a Lex_f_b Lex_f_c Lex_f_s CodeOKV)))))

(define ds-goods-price-fill
  (quasiquote
   (dataset
    (extend
     (GoodsRef
      (mode raw)
      (type Reference)
      (raw "(SELECT FIRST 1 g.\"fObject\" FROM \"tGoods\" g WHERE g.\"fDoc\" = :aDoc AND g.\"fModel\" = r_Goods.\"fModel\")"))
     (Inserted
      (mode raw)
      (type Boolean)
      (raw "(CASE WHEN EXISTS (SELECT g.\"fObject\" FROM \"tGoods\" g WHERE g.\"fDoc\" = :aDoc AND g.\"fModel\" = r_Goods.\"fModel\") THEN '1' ELSE '0' END)")
      )
     (XModelName
      (mode raw)
      (type LongString)
      (raw "CASE WHEN STRLEN(r_Goods.\"fQName\") > 0 THEN r_Goods.\"fQName\" ELSE r_Goods_Model.\"fName\" END")
      ))
    (fields (Goods OrderA Doc Count (Volume Name) Price NDS NDSMode (Model Name Curr (Volume Name) (Group Name)) Country GTD QName
                   GoodsRef Inserted XModelName))
    (ignore-ref Doc)
    (transform GoodsCost GoodsName)
    (where (Doc = :aPriceDoc)
           [nnvl ModelGroup = :aAutoFilter_ModelGroup]
           ))))

(define ds-dict
  (quasiquote
   (dataset
    (fields (Doc Type Name Firm)))))

(define ds-model-min
  (quasiquote
   (dataset
    (fields (Model Name)))))

(define ds-model-group-filter
  (quasiquote
   (dataset
    (fields (Model Name Group))
    (where [Name >= "''"]
           [nnvl Group = :aGroup]))))

(define ds-curr-min
  (quasiquote
   (dataset
    (fields (Curr Name)))))

(define ds-curr-lex
  (quasiquote
   (dataset
    (fields (Curr Name Lex_i_a Lex_i_b Lex_i_c Lex_i_s Lex_f_a Lex_f_b Lex_f_c Lex_f_s)))))

(define ds-doc-curr
  (quasiquote
   (dataset
    (fields (DocCurr Doc (Curr Name) Rate))
    (where (Doc = :aDoc)))))

(define ds-firm-min
  (quasiquote
   (dataset
    (fields (Firm ShortName Self))
    (where (Self = :aSelf)))))

(define ds-partner-min
  (quasiquote
   (dataset
    (fields (Firm ShortName Self)))))

(define ds-goods-min
  (quasiquote
   (dataset
    (fields (Goods OrderA Doc Count Price NDS NDSMode (Model Name Curr (Volume Name)) QName))
    (ignore-ref Doc)
    (where (Doc = :aDoc)))))

(define ds-goods-arch
  (quasiquote
   (dataset
    (extend
     (GoodsName
      (mode raw)
      (type LongString)
      (raw "(CASE WHEN r_Goods.\"fQName\" > '' THEN r_Goods.\"fQName\" ELSE r_Goods_Model.\"fName\" END)")))
    (fields (Goods OrderA Doc Count Price NDS NDSMode (Volume Name) (Model Name Curr) QName GoodsName))
    (ignore-ref Doc)
    (where (Doc = :aDoc)))))

(define ds-payment-std
  (quasiquote
   (dataset
    (fields (Payment Doc Desc Date Sum Curr))
    (ignore-ref Doc)
    (where (Doc = :aDoc)))))

(define ds-payment-min
  (quasiquote
   (dataset
    (fields (Payment Doc Desc Date Sum Curr))
    (ignore-ref Doc))))

(define ds-arch
  (quasiquote
   (dataset
    (extend
     (From
      (mode extend)
      (index 0)
      (type DateTime)
      (caption "с"))
     (To
      (mode extend)
      (index 1)
      (type DateTime)
      (caption "по"))
     (PayState
      (mode extend)
      (index 2)
      (type ImageName)
      (caption "Оплаты")))
    (fields (Doc Type Firm Partner Date Name From To PayState)))))

(define (ds-arch-docs type)
  (quasiquote
   (dataset
    (extend
     (PayState
      (mode raw)
      (type ImageName)
      (caption "PayState")
      (raw "(SELECT \"oState\" FROM \"pDocGetPaymentState\"(r_Doc.\"fObject\"))"))
     (PaySum
      (mode raw)
      (type Money)
      (caption "PaySum")
      ;(raw "COALESCE((SELECT SUM(\"fSum\") FROM \"tPayment\" WHERE \"fDoc\" = r_Doc.\"fObject\"), 0.0)")))
      (raw "(SELECT SUM(\"fSum\") FROM \"tPayment\" WHERE \"fDoc\" = r_Doc.\"fObject\")")))
    (fields (Doc Type (Firm ShortName) (Partner ShortName) Date Name Num Sum PaySum PayState Mark))
    (where (Type = ,(mix "'" type "'"))
           [nnvl Firm = :aFirm]
           [nnvl Partner = :aPartner]
           [nnvl Date >= :aFrom]
           [nnvl Date <= :aTo]
           [nnvl :aPayState CONTAINING PayState]))))

(define (ds-doc-min type)
  (quasiquote
   (dataset
    (fields (Doc Name Num Type))
    (where (Type = ,(mix "'" type "'"))))))

(define ds-attach
  (quasiquote
   (dataset
    (fields (Attach Owner Class Name Size State))
    (where [Owner = :aOwner]
           [Class = :aClass]))))

(define ds-subdoc-std
  (quasiquote
   (dataset
    (fields (SubDoc Doc Sum NDS NDSMode TypeAP OrderPL VidPL Purpose VidOsn NumOsn DateOsn Osn 
                    Partner PartnerCargo PayStatus KDOH OKATO OsnPL NalPer NumDok DateDok TypePL 
                    CodeIstKBK TypeKBKPay KBKPay TypeKBKRcp KBKRcp PurposeKBK NoteKBK OrderA
                    AddKlassPay AddKlassRcp))
    (aggregates
     (AggSum2 Sum Sum))
    (transform SubDocCost)
    (where (Doc = :aDoc)))))

(define ds-subdoc
  (quasiquote
   (dataset
    (fields (SubDoc Doc Sum NDS NDSMode TypeAP OrderPL VidPL Purpose VidOsn NumOsn DateOsn Osn 
                    Partner PartnerCargo PayStatus KDOH OKATO OsnPL NalPer NumDok DateDok TypePL 
                    CodeIstKBK TypeKBKPay KBKPay TypeKBKRcp KBKRcp PurposeKBK NoteKBK OrderA
                    AddKlassPay AddKlassRcp)))))

(define ds-expense-std
  (quasiquote
   (dataset
    (fields (Expense Doc Date Number Name SumVN SumRN DinV CinV OrderA))
    (aggregates
     (AggSumRN SumRN Sum))
    (where (Doc = :aDoc)))))

(define ds-expense-form
  (quasiquote
   (dataset
    (fields (Expense Doc Date Number Name SumVN SumRN DinV CinV OrderA)))))

(define ds-mutualref-ctrl
  (quasiquote
   (dataset
    (fields (MutualRef Owner0 Owner1 Name Value))
    (where (Owner0 = :aOwner0) (Owner1 = :aOwner1)))))

(define ds-okato-min
  (quasiquote
   (dataset
    (fields (Okato Name)))))

(define ds-kbk-min
  (quasiquote
   (dataset
    (fields (KBK Name TypeKBK))
    (where (TypeKBK = :aTypeKBK)))))

(define ds-kbk-inf
  (quasiquote
   (dataset
    (extend
     (NDesc
      (mode raw)
      (type LongString)
      (raw "CASE WHEN STRLEN(r_KBK.\"fDesc\") > 0 THEN r_KBK.\"fName\" || ' (' || r_KBK.\"fDesc\" || ')' ELSE r_KBK.\"fName\" END")))
    (fields (KBK NDesc Name TypeKBK))
    (where (TypeKBK = :aTypeKBK)))))

(define ds-kbk-std
  (quasiquote
   (dataset
    (fields (KBK Name TypeKBK Desc)))))

(define (ds-bank-find name)
  (quasiquote
   (dataset
    (fields (Bank Name BIK CorDep))
    (where ("UPPER(" ,name ")" "STARTING WITH" :aText)))))

(define ds-car-std
  (quasiquote
   (dataset
    (fields (Car Firm StateNum Mark GarageNum Type Model ReleaseYear Engine Chassis 
                 Body VIN Color PspSer PspNum CertSer CertNum CertDate CertGive)))))

(define ds-car-min
  (quasiquote
   (dataset
    (fields (Car StateNum Mark Type Firm))
    (where (Firm = :aFirm)
           (Type = :aType)))))

(define ds-goodsc-std
  (quasiquote
   (dataset
    (fields (GoodsC OrderA Doc Name Note PackType PackNum MassMode Code Container Class Mass))
    (ignore-ref Doc)
    (where (Doc = :aDoc)))))

(define ds-form-goodsc
  (quasiquote
   (dataset
    (ignore-ref Doc)
    (fields (GoodsC OrderA Doc Name Note PackType PackNum MassMode Code Container Class Mass)))))

(define ds-calc-method
  (quasiquote
   (dataset
    (fields (CalcMethod Name Method)))))
