(module dataset
  mzscheme
  
  (require
   "list.ss"
   (prefix u. "utils.scm")
   (prefix meta. "meta.scm")
   (prefix al. "alists.scm")
   (prefix df. "df-lists.scm")
   (prefix ms. "meta-sql.scm")
   "utils.scm")
  
  (provide q-field-params
           q-master
           q-fields
           q-select
           q-update
           q-update-x
           q-insert
           q-insert-x
           q-transform
           q-ignore-ref
           q-aggregates
           q-range-cond
           get-fields
           table-extend
           required-fields)
  
  (define meta meta.raw_meta)
  
  (define (d-f-m proc dfields pad tables)
    (if (null? dfields)
        (if (null? pad)
            null
            (d-f-m proc (car pad) (cdr pad) (cdr tables)))
        (let ([head (car dfields)]
              [tail (cdr dfields)])
          (if (list? head)
              (let ([table (car head)]
                    [rest (cdr head)])
                (cons (proc table tables) (d-f-m proc rest (cons tail pad) (cons table tables))))
              (cons (proc head tables) (d-f-m proc tail pad tables))))))
  
  (define (dfields-flat-map proc dfields)
    (let ([table (list (car dfields))]
          [rest (cdr dfields)])
      (cons (proc 'Object table) (d-f-m proc rest null table))))
  
  ; same as dfields-flat-map, but do not prepend Object field
  (define (dfields-flat-map* proc dfields)
    (let ([table (list (car dfields))]
          [rest (cdr dfields)])
      (d-f-m proc rest null table)))
  
  (define (d-t-l proc dfields pad tables)
    (if (null? dfields)
        (if (null? pad)
            null
            (d-t-l proc (car pad) (cdr pad) (cdr tables)))
        (let ([head (car dfields)]
              [tail (cdr dfields)])
          (if (list? head)
              (let ([table (car head)]
                    [rest (cdr head)])
                (cons (proc (cons table tables)) (d-t-l proc rest (cons tail pad) (cons table tables))))
              (d-t-l proc tail pad tables)))))
  
  (define (dfields-tables-list proc dfields)
    (let ([table (car dfields)]
          [rest (cdr dfields)])
      (d-t-l proc rest null (list table))))
  ;(d-t-l proc rest null null)))
  
  (define (dfields-master-map proc dfields)
    (let ([table (car dfields)]
          [rest (cdr dfields)])
      (map (lambda (field) (proc table (if (list? field) (car field) field))) rest)))
  
  (define (format-update-row table field)
    (format "  r_~a.\"f~a\" = :a~a" table field field))
  
  (define (format-insert-field table field)
    (format "\"f~a\"" field))
  
  (define (format-insert-param table field)
    (format ":a~a" field))
  
  (define (ds-list ds)
    (let* ([dfields (al.val1 ds 'fields)]
           [table (car dfields)]
           [fields (cdr dfields)])
      
      (define (build f) 
        (if (null? f)
            null
            (let [(head (car f))
                  (tail (cdr f))]
              (if (list? head)
                  (cons f f)
                  (cons head (build tail))
                  ))))
      
      fields))
  
  (define (same . v)
    (format "<~a>" v))
  
  (define (ds-func-factory dataset func)
    
    (define (format-tables-aliases tables)
      (if (null? tables)
          "r"
          (string-append (format-tables-aliases (cdr tables)) (format "_~a" (car tables)))))
    
    (define (format-field-alias field tables)
      
      (define (f-f-a tables)
        (if (or (null? tables) (null? (cdr tables)))
            ""
            (string-append (f-f-a (cdr tables)) (format "~a" (car tables)))))
      
      (format "f~a~a" (f-f-a tables) field))
    
    (define (format-field-name field tables)
      
      (define (f-f-a tables)
        (if (or (null? tables) (null? (cdr tables)))
            ""
            (string-append (f-f-a (cdr tables)) (format "~a" (car tables)))))
      
      (format "~a~a" (f-f-a tables) field))
    
    (define (f-f-extend table-alias field-alias spec) ;format-field-extend
      (let* ([type (al.val1 spec 'type)]
             [index (al.val1 spec 'index)]
             [ddl (ms.type-ddl type)]
             [mode (al.val1 spec 'mode)])
        (if (eq? mode 'raw)
            (mix "  CAST(" (al.val1 spec 'raw) " AS " ddl ") \"" field-alias "\"")
            (format "  CAST(NULLIF(~a_X.\"fV~a\", '') AS ~a) \"~a\"" table-alias index ddl field-alias))))
    
    (define (field-select-ddl f tables)
      (let ([spec (al.node dataset 'extend f)]
            [table-alias (format-tables-aliases tables)]
            [field-alias (format-field-alias f tables)])
        (if spec
            (f-f-extend table-alias field-alias spec)
            (format "  ~a.\"f~a\" \"~a\"" table-alias f field-alias)))
      )
    
    (define (table-name-std tables)
      (let* ([dfields (al.val1 dataset 'fields)]
             [rtables (reverse tables)]
             [meta-table (al.node meta 'tables (car rtables))])
        
        (define (t-n dfields tables-i meta-table)
          (if (null? tables-i)
              (car meta-table)
              (let* ([field (car tables-i)]
                     [type (al.val1 meta-table 'fields field 'type)]
                     [sub-mt (al.node meta 'tables type)]
                     [sub-df (df.get dfields field)])
                (t-n sub-df (cdr tables-i) sub-mt))))
        
        (t-n dfields (cdr rtables) meta-table)))
    
    (define (table-name tables)
      (if (and (= (length tables) 2) (al.val dataset 'extend (car tables)))
          (al.val1 dataset 'extend (car tables) 'type)
          (table-name-std tables)))
    
    (define (f-e-normal f tables table-alias field-alias spec) ;format-expr-normal
      (let* ([table (table-name tables)]
             [type-f (al.val1 meta 'tables table 'fields f 'type)]
             [type (if type-f type-f 'Reference)]
             [ddl1 (al.val1 meta 'types type 'ddl)]
             [case-sens (if (eq? ddl1 'VARCHAR) 'nocase 'case)]
             [expr (format "~a.\"f~a\"" table-alias f)])
        (list expr field-alias case-sens 'normal)))
    
    (define (f-e-extend table-alias field-alias spec) ;format-expr-extend
      (let* ([type (al.val1 spec 'type)]
             [index (al.val1 spec 'index)]
             [ddl (ms.type-ddl type)]
             [ddl1 (al.val1 meta 'types type 'ddl)]
             [case-sens (if (eq? ddl1 'VARCHAR) 'nocase 'case)]
             [expr (format "CAST(NULLIF(~a_X.\"fV~a\", '') AS ~a)" table-alias index ddl)])
        (list expr field-alias case-sens 'extend ddl)))
    
    (define (f-e-raw table-alias field-alias spec raw) ;format-expr-raw
      (let* ([type (al.val1 spec 'type)]
             [ddl (ms.type-ddl type)]
             [ddl1 (al.val1 meta 'types type 'ddl)]
             [case-sens (if (eq? ddl1 'VARCHAR) 'nocase 'case)]
             [expr (mix "CAST(" raw " AS " ddl ")")])
        (list expr field-alias case-sens 'extend ddl)))
    
    (define (field-expr f tables)
      (let ([spec (al.node dataset 'extend f)]
            [table-alias (format-tables-aliases tables)]
            [field-alias (format-field-alias f tables)])
        (if spec
            (let ([raw (al.val1 spec 'raw)])
              (if raw
                  (f-e-raw table-alias field-alias spec raw)
                  (f-e-extend table-alias field-alias spec)))
            (f-e-normal f tables table-alias field-alias spec))))
    
    (define (f-p-normal f tables table-alias field-alias spec) ;format-params-normal
      (let* ([table (table-name tables)]
             [caption (al.val1 meta 'tables table 'fields f 'caption)]
             [type-f (al.val1 meta 'tables table 'fields f 'type)]
             [type (if type-f type-f 'Reference)]
             [ddl1 (al.val1 meta 'types type 'ddl)]
             [case-sens (if (eq? ddl1 'VARCHAR) 'nocase 'case)]
             [expr (format "~a.\"f~a\"" table-alias f)])
        (list field-alias type caption)))
    
    (define (f-p-extend table-alias field-alias spec) ;format-params-extend
      (let* ([type (al.val1 spec 'type)]
             [index (al.val1 spec 'index)]
             [caption (al.val1 spec 'caption)]
             [ddl (ms.type-ddl type)]
             [ddl1 (al.val1 meta 'types type 'ddl)]
             [case-sens (if (eq? ddl1 'VARCHAR) 'nocase 'case)]
             [expr (format "CAST(NULLIF(~a_X.\"fV~a\", '') AS ~a)" table-alias index ddl)])
        (list field-alias type caption)))
    
    (define (field-params f tables)
      (let ([spec (al.node dataset 'extend f)]
            [table-alias (format-tables-aliases tables)]
            [field-alias (format-field-name f tables)])
        (if spec
            (f-p-extend table-alias field-alias spec)
            (f-p-normal f tables table-alias field-alias spec)
            ;(list (format "~a.\"f~a\"" (format-tables-aliases tables) f) field-alias 'normal)
            )))
    
    (define (get-field f tables)
      (let* ([spec (al.node dataset 'extend f)]
             [table (table-name tables)]
             [meta-field (if spec (cdr spec) (al.val meta 'tables table 'fields f))]
             [field-alias (format-field-name f tables)]
             [dname (string->symbol field-alias)])
        (if meta-field
            (cons dname (cons `(meta-name ,f) (cons `(table ,table) meta-field)))
            (list dname))))
    
    (define (extend-join-expr m-alias tables)
      (let* ([field (car tables)]
             [index (al.val1 dataset 'extend field 'index)])
        (mix "COALESCE(NULLIF(CAST(RTRIM(SUBSTR(r_Doc_X.\"fV" index "\", 1, 12)) AS CHAR(12)), ''), '000000000000')")))
    
    (define (table-join-dll tables)
      (let* ([slave (car tables)]
             [master (cadr tables)]
             [s-table (format "\"t~a\"" (table-name tables))]
             [s-alias (format-tables-aliases tables)]
             [m-alias (format-tables-aliases (cdr tables))]
             [j-expr (if (and (= (length tables) 2) (al.val dataset 'extend (car tables)))
                         (extend-join-expr m-alias tables)
                         (format "~a.\"f~a\"" m-alias slave))])
        (format "[j]JOIN\n  ~a ~a ON (~a.\"fObject\" = ~a)" s-table s-alias s-alias j-expr)))
    
    (case func
      ((format-tables-aliases) format-tables-aliases)
      ((format-field-alias) format-field-alias)
      ((field-select-ddl) field-select-ddl)
      ((field-expr) field-expr)
      ((field-params) field-params)
      ((get-field) get-field)
      ((table-join-dll) table-join-dll)
      (else (error "unknown function"))))
  
  (define macro-names '(nnvl mtest))
  
  (define (find-field qfields field)
    (let* ([sname (format "f~a" field)]
           [qfield (filter (lambda (f) (equal? (list-ref f 1) sname)) qfields)])
      (if (null? qfield)
          #f
          (car qfield))))
  
  (define (format-where-expr dataset)
    (let ([where (al.val dataset 'where)])
      (if where
          (let ([qfields (q-fields dataset)])
            
            (define (expand-one e)
              (let ([s (format "~a" e)]
                    [qfield (find-field qfields e)])
                (cond
                  ((string? e)
                   e)
                  (qfield
                   (car qfield))
                  ((list? e)
                   (expand-where e))
                  (else
                   s))))
            
            (define (expand-macro m)
              (let ([name (car m)]
                    [rest (cdr m)])
                (list " " (format "[~a=~a]" name (expand-expr rest)))))
            
            (define (expand-expr w)
              (if (member (car w) macro-names)
                  (expand-macro w)
                  (u.concat-str (map expand-one w) " ")))
            
            (define (expand-where w)
              (if (member (car w) macro-names)
                  (expand-macro w)
                  (list " and " (u.mix "(" (u.concat-str (map expand-one w) " ") ")"))))
            
            (u.mix1 (cdr (u.flatten (map expand-where where)))))
          "")))
  
  (define (format-where dataset)
    (let ([where (al.val dataset 'where)])
      (if where
          (u.mix "WHERE\n  " (format-where-expr dataset) "\n  [and-where]")
          "[where]")))
  
  (define (format-where.b00 dataset)
    (let ([where (al.val dataset 'where)])
      (if where
          (let ([qfields (q-fields dataset)])
            
            (define (expand-one e)
              (let ([s (format "~a" e)]
                    [qfield (find-field qfields e)])
                (cond
                  ((string? e)
                   e)
                  (qfield
                   (car qfield))
                  ((list? e)
                   (expand-where e))
                  (else
                   s))))
            
            (define (expand-macro m)
              (let ([name (car m)]
                    [rest (cdr m)])
                (list " " (format "[~a=~a]" name (expand-expr rest)))))
            
            (define (expand-expr w)
              (if (member (car w) macro-names)
                  (expand-macro w)
                  (u.concat-str (map expand-one w) " ")))
            
            (define (expand-where w)
              (if (member (car w) macro-names)
                  (expand-macro w)
                  (list " and " (u.mix "(" (u.concat-str (map expand-one w) " ") ")"))))
            
            (u.mix "WHERE\n  " (u.mix1 (cdr (u.flatten (map expand-where where)))) "\n  [and-where]"))
          "[where]")))
  
  (define (q-master dataset)
    (car (al.val1 dataset 'fields)))
  
  (define (q-select dataset)
    (let* ([dfields (al.val1 dataset 'fields)]
           [master (car dfields)]
           [m-table (format "\"t~a\"" master)]
           [m-alias (format "r_~a" master)]
           [s-fields (u.concat-str (dfields-flat-map (ds-func-factory dataset 'field-select-ddl) dfields) ",\n")]
           [s-joins (u.concat-str (dfields-tables-list (ds-func-factory dataset 'table-join-dll) dfields) "\n")]
           [e-join (if (al.node dataset 'extend)
                       (format "[j]JOIN\n  \"tExtend\" ~a_X on (~a_X.\"fOwner\" = CASE WHEN EXISTS (SELECT 1 FROM \"tExtend\" WHERE \"fOwner\" = ~a.\"fObject\") THEN ~a.\"fObject\" ELSE '000000000000' END)\n" m-alias m-alias m-alias m-alias)
                       "")]
           [s-where (format-where dataset)])
      (string-append
       "SELECT [first]\n"
       s-fields "[fields]\n"
       "FROM[f]\n"
       (format "  ~a ~a [master]\n" m-table m-alias)
       e-join
       s-joins "\n"
       s-where "\n"
       "[order]"
       )))
  
  (define (q-range-cond dataset)
    (let* ([dfields (al.val1 dataset 'fields)]
           [master (car dfields)]
           [m-table (format "\"t~a\"" master)]
           [m-alias (format "r_~a" master)]
           [s-where (format-where-expr dataset)])
      (list m-table m-alias s-where)))

  (define (q-fields dataset)
    (let* ([dfields (al.val1 dataset 'fields)]
           [s-fields (dfields-flat-map (ds-func-factory dataset 'field-expr) dfields)])
      s-fields))
  
  (define (q-field-params dataset)
    (let ([dfields (al.val1 dataset 'fields)])
      (dfields-flat-map (ds-func-factory dataset 'field-params) dfields)))
  
  (define (get-fields dataset)
    (let ([dfields (al.val1 dataset 'fields)])
      (cons 'fields (dfields-flat-map* (ds-func-factory dataset 'get-field) dfields))))
  
  (define (field-normal? dataset field)
    (if (al.node dataset 'extend field)
        #f
        #t))
  
  (define (field-extend? dataset field)
    (let ([spec (al.node dataset 'extend field)])
      (if spec
          (eq? (al.val1 spec 'mode) 'extend)
          #f)))
  
  (define (field-normal-test dataset)
    (lambda (field)
      (field-normal? dataset field)))
  
  (define (field-extend-test dataset)
    (lambda (field)
      (field-extend? dataset field)))
  
  (define (invert proc)
    (lambda params
      (not (apply proc params))))
  
  (define (dataset-master-base-fields dataset)
    (let* ([dfields (al.val1 dataset 'fields)]
           [base (df.base-fields dfields)])
      (filter (field-normal-test dataset) base)))
  
  (define (dataset-master-extend-fields dataset)
    (let* ([dfields (al.val1 dataset 'fields)]
           [base (df.base-fields dfields)])
      (filter (field-extend-test dataset) base)))
  
  (define (fmt-update-row table)
    (lambda (field)
      (format "  r_~a.\"f~a\" = :a~a" table field field)))
  
  (define (q-update dataset)
    (let* ([dfields (al.val1 dataset 'fields)]
           [master (car dfields)]
           [master-base (dataset-master-base-fields dataset)]
           [s-fields (u.concat-str (map (fmt-update-row master) master-base) ",\n")])
      (string-append
       "UPDATE\n"
       (format "  \"t~a\" r_~a\n" master master)
       "SET\n"
       s-fields "\n"
       "WHERE\n"
       (format "  r_~a.\"fObject\" = :aObject" master)
       )))
  
  (define (fmt-update-row-x dataset)
    (lambda (field)
      (let* ([extend (al.node dataset 'extend field)]
             [index (al.val1 extend 'index)]
             [type (al.val1 extend 'type)]
             [ddl (ms.type-ddl type)])
        (format "  x.\"fV~a\" = CAST(COALESCE(CAST(NULL AS ~a), :a~a) AS VARCHAR(750))" index ddl field))))
  
  (define (has-extend? dataset)
    (if (al.node dataset 'extend)
        #t
        #f))
  
  (define (q-update-x dataset)
    (if (has-extend? dataset)
        (let* ([dfields (al.val1 dataset 'fields)]
               [master (car dfields)]
               [master-base (dataset-master-extend-fields dataset)]
               [s-fields (u.concat-str (map (fmt-update-row-x dataset) master-base) ",\n")])
          (if (null? master-base)
              ""
              (string-append
               "UPDATE\n"
               "  \"tExtend\" x\n"
               "SET\n"
               s-fields "\n"
               "WHERE\n"
               "  x.\"fOwner\" = :aObject"
               )))
        ""
        ))
  
  (define (fmt-insert-field table)
    (lambda (field)
      (format "\"f~a\"" field)))
  
  (define (fmt-insert-param table)
    (lambda (field)
      (format ":a~a" field)))
  
  (define (q-insert dataset)
    (let* ([dfields (al.val1 dataset 'fields)]
           [master (car dfields)]
           [master-base (dataset-master-base-fields dataset)]
           [s-fields (u.concat-str (map (fmt-insert-field master) master-base) ", ")]
           [s-params (u.concat-str (map (fmt-insert-param master) master-base) ", ")])
      (string-append
       "INSERT INTO " (format "\"t~a\"\n" master)
       "  (\"fObject\", " s-fields ")\n"
       "VALUES\n"
       "  (:aObject, " s-params ")"
       )))
  
  (define (fmt-insert-field-x dataset)
    (lambda (field)
      (let* ([extend (al.node dataset 'extend field)]
             [index (al.val1 extend 'index)])
        (format "\"fV~a\"" index))))
  
  (define (fmt-insert-param-x dataset)
    (lambda (field)
      (let* ([extend (al.node dataset 'extend field)]
             [index (al.val1 extend 'index)]
             [type (al.val1 extend 'type)]
             [ddl (ms.type-ddl type)])
        (format "CAST(COALESCE(CAST(NULL AS ~a), :a~a) AS VARCHAR(200))" ddl field))))
  
  (define (q-insert-x dataset)
    (if (has-extend? dataset)
        (let* ([dfields (al.val1 dataset 'fields)]
               [master (car dfields)]
               [master-base (dataset-master-extend-fields dataset)]
               [s-fields (u.concat-str (map (fmt-insert-field-x dataset) master-base) ", ")]
               [s-params (u.concat-str (map (fmt-insert-param-x dataset) master-base) ", ")])
          (if (null? master-base)
              ""
              (string-append
               "INSERT INTO \"tExtend\"\n"
               "  (\"fObject\", \"fOwner\", " s-fields ")\n"
               "VALUES\n"
               "  (:aExtend, :aObject, " s-params ")"
               )))
        ""))
  
  (define (q-transform dataset)
    (let ([transform (al.val dataset 'transform)])
      (if transform
          transform
          null)))
  
  (define (q-ignore-ref dataset)
    (let ([ignore-ref (al.val dataset 'ignore-ref)])
      (if ignore-ref
          ignore-ref
          null)))
  
  (define (q-aggregates dataset)
    (let ([aggregates (al.val dataset 'aggregates)])
      (if aggregates
          aggregates
          null)))
  
  (define (flat-extend field)
    (let* ([len (length field)]
           [pn (lambda (p n) (if (< n len) (list p (list-ref field n)) #f))]
           [name (list-ref field 0)]
           [tags (if (> (length field) 4) (cons 'tags (list-ref field 4)) #f)]
           [params (filter list? (list '(mode extend)
                                       (pn 'index 1)
                                       (pn 'type 2)
                                       (pn 'caption 3)
                                       tags))])
      (cons name params)))
  
  (define (table-extend table ext)
    (let* ([extend (map flat-extend ext)]
           [e-fields (map car extend)]
           [fields (append (map car (al.val meta 'tables 'Custom 'fields)) e-fields)]
           [d-fields (cons table fields)])
      (quasiquote
       (dataset
        ,(cons 'extend extend)
        (fields ,d-fields)
        ))))
  
  (define (base-field field)
    (if (list? field)
        (car field)
        field))
  
  ; at now supports only owner fields
  (define (required-fields ds)
    (let* ([dfields (al.val1 ds 'fields)]
           [table-n (car dfields)]
           [table-fields (al.node meta 'tables table-n 'fields)]
           [fields (cdr dfields)]
           [base-fields (map base-field fields)]
           [owner-fields (filter (lambda (field)
                                   (let ([tags (al.val table-fields field 'tags)])
                                     (if tags
                                         (member 'owner tags)
                                         #f)))
                                 base-fields)])
      owner-fields))
  
  
  )

#|
(require dataset)

(define export-1c-a.ds
  (quasiquote
   (dataset
    (extend
     (PayType
      (mode extend)
      (index 0)
      (type ShortString)
      (caption "Вид платежа"))
     (State
      (mode extend)
      (index 1)
      (type ShortString)
      (caption "Статус"))
     (PayKind
      (mode extend)
      (index 2)
      (type ShortString))
     (PayDate
      (mode extend)
      (index 3)
      (type Date))
     (PayOrd
      (mode extend)
      (index 4)
      (type ShortString))
     (PayCode
      (mode extend)
      (index 5)
      (type ShortString))
     (PClass
      (mode extend)
      (index 6)
      (type KBK))
     (POKAT
      (mode extend)
      (index 7)
      (type Okato))
     (PBase
      (mode extend)
      (index 8)
      (type ShortString))
     (PPeriod
      (mode extend)
      (index 9)
      (type ShortString))
     (PNum
      (mode extend)
      (index 10)
      (type ShortString))
     (PDocDt
      (mode extend)
      (index 11)
      (type ShortString))
     (PKind
      (mode extend)
      (index 12)
      (type ShortString))
     (Purpose
      (mode extend)
      (index 13)
      (type MemoString))
     (NDS
      (mode extend)
      (index 14)
      (type Percent)
      (caption "НДС"))
     (No2ndSign
      (mode extend)
      (index 15)
      (type LongString))
     (NoNDS
      (mode extend)
      (index 16)
      (type LongString))
     (NDSStr
      (mode extend)
      (index 17)
      (type LongString)))
    (fields (Doc Type Name Date PayType PayKind PayDate PayOrd Purpose NDSStr State
                 (PClass Name) 
                 ;PClass
                 POKAT PBase PPeriod PNum PDocDt PKind
                 (Firm FullName INN KPP OfkOrg OfkDep OfkCode OfkINN Bank BIK RasDep CorDep)
                 (Partner FullName INN KPP OfkOrg OfkDep OfkCode OfkINN Bank BIK RasDep CorDep) Sum)))))

(define ds-goods-std
  (quasiquote
   (dataset
    (fields (Goods OrderA Doc Count Price NDS NDSMode (Model Name Curr (Volume Name) )  QName))
    )))

(display (q-select export-1c-a.ds))
;(q-select ds-goods-std)
|#