(require "list.ss"
         (prefix al. "alists.scm")
         "utils.scm"
         "meta-utils3.scm"         ; do not remove - used in TpvFBManager.GetOwnerOfObject
         (prefix meta. "meta.scm") ;         
         )

(load "common-defs.scm")

(define meta meta.raw_meta)

(define scm-suffix (regexp "\\.scm$"))

(define (file-is-doc file)
  (regexp-match scm-suffix (path->string file)))

(define (doc-dir base) (build-path root-path base))

(define (import-form base)
  (lambda (name)
    (let ([path (build-path (doc-dir base) name)])
      (load path)
      (let ([e (safe-eval 'export)])
        (if (and e (list? e))
            (al.val e 'forms)
            #f)))))

(define (refine-import import base)
  (apply append (filter (lambda (x) x) import)))

(define (document-forms base)
  (let ([dir (doc-dir base)])
    (if (directory-exists? dir)
        (refine-import (map (import-form base) (filter file-is-doc (directory-list dir))) base)
        null)))

(define (document-info base)
  (map (lambda (name)
         (let* ([form (eval name)]
                [caption (al.val1 form 'caption)]
                [archive (al.val1 form 'archive)])
           (list name caption archive))) (document-forms base)))

(define (export-info base)
  (map (lambda (name)
         (let* ([form (eval name)]
                [caption (al.val1 form 'caption)]
                [archive (al.val1 form 'export)])
           (list name caption archive))) (document-forms base)))

(define default-document-info
  (append (document-info "documents")
          (document-info "documents.extend")))
(define default-export-info (export-info "exports"))

(define (doc-fill-caption doc-info)
  (let* ([name (car doc-info)]
         [form (eval name)]
         [caption (al.val1 form 'fill 'caption)]
         [arch (caddr doc-info)])
    (list name caption arch)))

(define (doc-fill-sufficient doc-name req)
  (let ([info (assoc doc-name default-document-info)])
    (if info
        (let* ([doc (eval (car info))]
               [provide (al.val doc 'fill 'provide)])
          (if provide
              (includes provide req)
              #f))
        #f)))

(define (doc-fill-source doc-name)
  (let ([info (assoc doc-name default-document-info)])
    (if info
        (let* ([doc (eval (car info))]
               [req (al.val doc 'fill 'require)])
          (if req
              (map doc-fill-caption
                   (filter (lambda (doc-info)
                             (and (not (eq? (car doc-info) doc-name))
                                  (doc-fill-sufficient (car doc-info) req)))
                           default-document-info))
              null))
        null)))

(define (req-fill-source req)
  (map doc-fill-caption
       (filter (lambda (doc-info)
                 (doc-fill-sufficient (car doc-info) req))
               default-document-info)))
