(module layout
  mzscheme
  
  (require
   "list.ss"
   "utils.scm"
   (prefix meta. "meta.scm")
   (prefix al. "alists.scm")
   (prefix ds. "dataset.scm"))
  
  (provide form-get-expand-layout
           form-get-childs
           form-get-caption
           form-get-expand-child-layout
           non-visual-controls-expand
           
           style-of-control)
  
  (define visual-defaults
    (quasiquote
     (defaults
       (Model
        (dataset ds-model-min)
        (form form-model-std))
       (Volume
        (dataset ds-volume-min)
        (form form-volume-std))
       (Curr
        (dataset ds-curr-std)
        (form form-curr-std))
       (WeakDocCurr
        (dataset ds-curr-std)
        (form form-curr-std))
       (People
        (dataset ds-people-min)
        (form form-people-std))
       (Dutie
        (dataset ds-dutie-min)
        (form form-dutie-std))
       (Country
        (dataset ds-country-min)
        (form form-country-std))
       (Group
        (dataset ds-group-min)
        (form form-group-std))
       ;GB
       (ModType
        (dataset ds-model-type-min)
        (form form-model-type-std))
       (Producer
        (dataset ds-producer-min)
        (form form-producer-std))
       (CondDlv
        (dataset ds-cond-dlv-min)
        (form form-cond-dlv-std))
       (DestDlv
        (dataset ds-dest-dlv-min)
        (form form-dest-dlv-std))
       )))
  
  (define meta meta.raw_meta)
  
  (define (map-iter proc l)
    
    (define (m-i l index)
      (if (null? l)
          null
          (cons (proc (car l) index) (m-i (cdr l) (+ index 1)))))
    
    (m-i l 0))
  
  (define (layout-map_ proc layout)
    
    (define (l-m proc layout index)
      (let* ([dim (car layout)]; dimension v h or t
             [tail (cdr layout)]
             [ctrl (if (eq? dim 't) (cdr tail) tail)]
             [head (if (eq? dim 't) (list dim (car tail)) (list dim))])
        (append head
                (map (lambda (item)
                       (if (list? item)
                           (l-m proc item (+ 1 index))
                           (proc dim index item)))
                     ctrl))))
    
    (l-m proc layout 0))
  
  (define (layout-map proc layout)
    
    (define (l-m proc layout index path)
      (let* ([dim (car layout)]; dimension v h or t
             [tail (cdr layout)]
             [ctrl (if (eq? dim 't) (cdr tail) tail)]
             [head (if (eq? dim 't) (list dim (car tail)) (list dim))])
        (append head
                (map-iter (lambda (item item-index)
                            (let ([new-path (cons (list dim item-index) path)])
                              (if (list? item)
                                  (l-m proc item (+ 1 index) new-path)
                                  (proc dim index item new-path))))
                          ctrl))))
    
    (l-m proc layout 0 null))
  
  (define (layout-get layout control)
    
    (define (l-g items)
      (if (null? items)
          #f
          (let ([item (car items)]
                [rest (cdr items)])
            (if (list? item)
                (let ([val (layout-get item control)])
                  (if val
                      val
                      (l-g rest)))
                (if (eq? item control)
                    item
                    (l-g rest)
                    )))))
    
    (let* ([type (car layout)]
           [items (if (eq? type 't) (cddr layout) (cdr layout))])
      (l-g items)))
  
  (define (non-visual-controls form)
    (let* ([layout (al.val1 form 'layout)]
           [f-c (al.val form 'controls)]
           [controls (if f-c (map car f-c) null)])
      (filter (lambda (item) (not (layout-get layout item))) controls)))
  
  (define (non-visual-controls-expand form)
    (map (lambda (item) ((expand-visual-proc form) 'h 0 item null)) (non-visual-controls form)))
  
  (define (find-group control groups)
    
    (define (f-g groups index)
      (if (null? groups)
          #f
          (if (member control (car groups))
              index
              (f-g (cdr groups) (+ 1 index))
              )))
    (if groups
        (f-g groups 0)
        #f))
  
  (define (visual-type type)
    (let* ([ddl (al.val meta 'types type 'ddl)]
           [ddl-s (if ddl
                      ddl 
                      (if (al.val meta 'tables type)
                          '(OBJECT)
                          '(UNKNOWN)))]
           [first (car ddl-s)])
      
      (case first
        ((OBJECT) 'Reference)
        ;((VARCHAR) 'String)
        ((VARCHAR) (if (> (cadr ddl-s) 1000) 'Memo 'String))
        ((CHAR) (case (cadr ddl-s)
                  ((1) 'Boolean)
                  ((12) 'Reference)
                  (else 'String)))
        ((INTEGER) 'Integer)
        ((DOUBLE) 'Float)
        ((DATE) 'DateTime)
        ((TIMESTAMP) 'DateTime)
        (else 'Unknown))))
  
  (define (max-len type)
    (let ([ddl (al.val meta 'types type 'ddl)])
      (if ddl
          (if (eq? (car ddl) 'VARCHAR)
              (cadr ddl)
              #f)
          #f)))
  
  (define (style-has-control style control)
    (let ([controls (cdr style)])
      (member control controls)))
  
  (define (style-of-control form control)
    (let ([styles (al.val1 form 'styles)])
      (if styles
          (let* ([contains (filter (lambda (style) (style-has-control style control)) styles)]
                 [names (map car contains)])
            (if (null? names)
                #f
                names;(car names)
                ))
          #f)))
  
  (define (expand-label form dim index control group style)
    (quasiquote
     (c
      (name free)
      (type Label)
      (group ,group)
      (caption ,control))))
  
  (define (expand-control form dim index control group style)
    (let* ([ctrl (al.val form 'controls control)]
           [caption (al.val1 ctrl 'caption)]
           [form-control (if style;**
                             (cons 'c (cons (cons 'style style) (cons (list 'name control) ctrl)))
                             (cons 'c (cons (list 'name control) ctrl)))])
      (if caption
          (quasiquote
           (h (c
               (name free)
               (type Label)
               (group ,group)
               (caption ,caption)
               (text-align right))
              ,form-control))
          form-control)))
  
  (define (expand-field form dataset fields dim index control group style)
    (let* ([table (car (al.val1 dataset 'fields))]
           [m-field (al.node fields control)]
           [tags (al.val m-field 'tags)]
           [f-type (al.val1 m-field 'type)]
           [type (visual-type f-type)]
           [m-l (max-len f-type)]
           [v-caption (al.val1 form 'visual control 'caption)]
           [caption (if v-caption (if (eq? v-caption 'none) #f v-caption) (al.val1 m-field 'caption))]
           [params-0 `((name ,control)
                       (type ,type)
                       (bind auto-name))]
           [params-1 (if (eq? type 'Reference) 
                         (let* ([visual (al.node form 'visual control)]
                                [v-data (al.val1 visual 'dataset)]
                                [v-form (al.val1 visual 'form)]
                                [r-data (if v-data v-data (al.val1 visual-defaults f-type 'dataset))]
                                [r-form (if v-form v-form (al.val1 visual-defaults f-type 'form))])
                           (cons `(dataset ,r-data) (cons `(form ,r-form) params-0)))
                         params-0)]
           [params-2 (if style
                         ;(cons `(style ,style) params-1);**
                         (cons (cons 'style style) params-1);**
                         params-1)]
           [params-3 (if m-l
                         (cons `(max-len ,m-l) params-2)
                         params-2)]
           [params-4 (if (and tags (member 'template tags))
                         (cons `(template ,(mix table "." control)) params-3)
                         params-3)]
           [params-5 (if (and tags (member 'select-file tags))
                         (cons `(select-file ,(mix table "." control)) params-4)
                         params-4)]
           [params-6 (if (and tags (member 'select-bank tags))
                         (cons `(select-bank ,(mix table "." control)) params-5)
                         params-5)]
           [params-7 (al.merge params-6 (al.val form 'visual control))]
           [field (cons 'c params-7)])
      (if caption
          (quasiquote
           (h (c
               (name free)
               (type Label)
               (group ,group)
               (caption ,caption)
               (text-align right))
              ,field))
          field)))
  
  (define (path->group path)
    (if (< (length path) 2)
        'free
        (let ([first (list-ref path 0)]
              [second (list-ref path 1)])
          (if (and (eq? (car first) 'h) (eq? (cadr first) '0))
              (mix "ag_" (apply mix (map-iter (lambda (item index) (if (eq? index 1) (mix (car item) "_") (mix (car item) (cadr item)))) path)))
              'free))))
  
  (define (expand-visual form dataset fields dim index control path)
    (let* ([controls (al.node form 'controls)]
           [groups (al.val1 form 'groups)]
           [form-group (find-group control groups)]
           [group (if form-group 
                      (format "f~a" form-group)
                      (if (eq? dim 'v)
                          (format "v~a" index)
                          'free)
                      ;(path->group path)
                      )]
           [style (style-of-control form control)]) ;!!
      
      (cond
        ((string? control) (expand-label form dim index control (if form-group group 'free) style))
        ((al.node controls control) (expand-control form dim index control group style))
        (else (expand-field form dataset fields dim index control group style)))))
  
  (define (expand-visual-proc form)
    (let* ([dataset (eval (al.val1 form 'dataset))]
           [fields (ds.get-fields dataset)])
      (lambda (dim index control path)
        (expand-visual form dataset fields dim index control path))))
  
  (define (form-get-expand-layout form)
    (let ([layout (al.val1 form 'layout)])
      (layout-map (expand-visual-proc form) layout)))
  
  (define (form-get-expand-child-layout form child)
    (let* ([child-val (al.val form 'childs child)]
           [layout (cadr child-val)])
      (layout-map (expand-visual-proc form) layout)))
  
  (define (form-get-caption form)
    (let ([caption (al.val1 form 'caption)])
      (if caption
          caption
          "")))
  
  (define (form-get-child-caption form child)
    (let* ([child-val (al.val form 'childs child)]
           [caption (car child-val)])
      caption))
  
  (define (form-get-childs form)
    (let ([childs (al.val form 'childs)])
      (if childs
          (map (lambda (child) (list (car child) (cadr child))) childs)
          null)))
  
  )
