(module meta-utils3
  mzscheme
  
  (provide field-owner?
           get-owner-field
           get-desc-field
           get-depends
           type-depends
           merge-over)
  
  (require "list.ss"
           (prefix al. "alists.scm"))
  
  (define (field-owner? field)
    (let ([tags (al.val field 'tags)])
      (if tags
          (if (or (member 'owner tags) (member 'ext-owner tags)) #t #f)
          #f)))
  
  (define (get-owner-field table)
    (let* ([fields (filter field-owner? (al.val table 'fields))])
      (if (null? fields)
          #f
          (car fields))))
  
  (define (get-desc-field table)
    (let* ([desc-fields (filter (lambda (field) (let ([tags (al.val field 'tags)]) (if tags (member 'desc tags) #f)))
                                (al.val table 'fields))]
           [desc-field (if (null? desc-fields) 'Object (caar desc-fields))])
      (if (null? desc-fields) 'Object (caar desc-fields))))
  
  (define (get-depends table tables)
    
    (define (g-d tables)
      (if (null? tables)
          null
          (let* ([dep-table (car tables)]
                 [rest (cdr tables)]
                 [dep-fields (filter (lambda (field)
                                       (let ([type (al.val1 field 'type)]
                                             [tags (al.val field 'tags)])
                                         (and 
                                          (eq? type table)
                                          (not (if tags (member 'owner tags) #f)))))
                                     (al.val dep-table 'fields))]
                 [dep-names (map car dep-fields)])
            (if (null? dep-names)
                (g-d rest)
                (let* ([table-name (car dep-table)]
                       [table-desc (al.val1 dep-table 'desc)]
                       [desc (if table-desc table-desc table-name)])
                  (cons (list table-name dep-names (get-desc-field dep-table) desc) (g-d rest)))))))
    
    (g-d tables))
  
  (define (type-depends tables type)
    (filter (lambda (dep) (not (null? (cdr dep))))
            (map (lambda (table)
                   (cons (car table)
                         (map (lambda (field)
                                (car field))
                              (filter (lambda (field)
                                        (eq? (al.val1 field 'type) type))
                                      (al.val table 'fields)))))
                 tables)))
  
  (define (merge-over meta meta-over)
    (let ([over (al.val meta-over 'over)]
          [version (al.val1 meta-over 'version)])
      (al.tree-map (lambda (item mode path)
                     (case mode
                       ((node) (if (equal? path '(meta version))
                                   (list 'version (list-ref item 1) (list-ref item 2) version (list-ref item 4))
                                   (let ([over-items (assoc path over)])
                                     (if over-items
                                         (let ([method (cadr over-items)]
                                               [content (cddr over-items)])
                                           (case method
                                             ((+) (append item content))
                                             ((=) (car content))
                                             (else (error "unsupported merge method"))))
                                         item))))
                       (else item)))
                   meta)))
  
  )