(module query-gen-adv
  mzscheme
  
  (require
   "list.ss"
   "utils.scm"
   (prefix al. "alists.scm")
   (prefix meta. "meta.scm"))
  
  (provide
   
   table-default-fields
   table-base-fields
   query-field-defs
   query-select
   query-refresh
   query-update
   query-insert
   query-delete
   fields-inf
   query-ref
   query-def
   query-owner
   
   table-visual-fields
   type-is-ref?
   type-ddl
   
   make-desc
   
   shallow-copy-record-proc-ddl
   
   ; debug
   restore-refs
   get-childs-of-table)
  
  (define t-proc 'procedure)
  
  (define (desc-print desc port write?)
    (write-string (symbol->string (desc-ref desc 0)) port))
  
  (define-values (s:desc make-desc desc? desc-ref desc-set!)
    (make-struct-type 'desc #f 2 0 #f
                      (list (cons prop:custom-write desc-print))))
  
  (define (desc-field desc)
    (desc-ref desc 0))
  
  (define (desc-str desc)
    (desc-ref desc 1))
  
  (define meta (meta.meta))
  
  (define (norm-key key)
    (if (desc? key)
        (desc-field key)
        key))
  
  (define (a-get alist v)
    (if (null? v)
        alist
        (if alist
            (a-get (assq (norm-key (car v)) (cdr alist)) (cdr v))
            #f)))
  
  (define (safe-get alist . v)
    (let [(val (a-get alist v))]
      (if val
          (cdr val)
          null)))
  
  (define (safe-get-val alist . v)
    (let [(val (a-get alist v))]
      (if val
          (cdr val)
          #f)))
  
  (define (safe-get-one alist . v)
    (let [(val (a-get alist v))]
      (if val
          (car (cdr val))
          #f)))
  
  ;(define br "\n")
  
  (define (fold-strings strings)
    (apply string-append strings))
  
  (define (fold-disp inp)
    (apply string-append (map (lambda (i) (format "~a" i)) inp)))
  
  (define (format-cond cnd)
    (format "~n  AND (~a)" cnd))
  
  (define (format-ord cnd)
    (format ", ~a" cnd))
  
  (define fmt-select-ext ",~n  CAST((SELECT \"fValue\" FROM \"tExt\" WHERE (\"fOwner\" = ~a) AND (\"fName\" = '~a')) AS ~a) \"f~a\"")
  
  (define fmt-field-def-ext "CAST((SELECT \"fValue\" FROM \"tExt\" WHERE (\"fOwner\" = ~a) AND (\"fName\" = '~a')) AS ~a)")
  
  (define fmt-field-inf "CAST((SELECT \"fValue\" FROM \"tExt\" WHERE (\"fOwner\" = ~a) AND (\"fName\" = '~a')) AS ~a)")
  
  (define (safe-cadr* a)
    (if (and (pair? a) (pair? (cdr a)))
        (cadr a)
        null))
  
  (define (plain-type-ddl type)
    (if (null? type)
        ""
        (let [(first (car type))
              (next (safe-cadr* type))]
          (string-append
           (cond
             ((and (symbol? first) (number? next))
              (format "~a(" first))
             ((and (number? first) (number? next))
              (format "~a, " first))
             ((and (number? first) (null? next))
              (format "~a)" first))
             ((and (symbol? first) (null? next))
              (format "~a" first))
             (else
              (format "~a " first)))
           (plain-type-ddl (cdr type))))))    
  
  (define (type-ddl type)
    (let* [(stype (if (list? type) (car type) type))
           (plain (safe-get-val meta 'types stype 'ddl))]
      (if plain
          (plain-type-ddl plain)
          (format "CHAR(12) /* ~a */" stype))))
  
  (define q0 '(Goods Count Price NDS (Model Name Price (Group Name PayType) (Curr Name)) PayType PayKind))
  ;(define q1 '(Goods Count Price NDS (Model Name Price )))
  
  (define (fields-hier fields)
    (define (hier inp out add)
      (if (null? inp)
          out
          (let [(first (car inp)) (rest (cdr inp))]
            (if (list? first)
                (hier rest (hier (cdr first) (cons (list (car first) add) out) (cons (car first) add)) add)
                (hier rest (cons (list first add) out) add)))))
    (map
     (lambda (item) (list (car item) (reverse (cadr item))))
     (reverse (hier (cdr fields) null (list (car fields))))))
  
  (define (join-hier fields)
    (define (hier inp out add)
      (if (null? inp)
          out
          (let [(first (car inp)) (rest (cdr inp))]
            (if (list? first)
                (hier rest (hier first (cons (cons (car first) add) out) (cons (car first) add)) add)
                (hier rest out add)))))
    (map reverse (reverse (hier (cdr fields) null (list (car fields))))))
  
  (define (field-type field)
    (let* [(name (car field))
           (tables (cadr field))
           (table (car tables))
           (rest (cdr tables))
           (fields (append rest (list name)))]
      
      (define (get table fields)
        (if (null? fields)
            table
            (let* [(name (car fields))
                   (rest (cdr fields))
                   (ftype (safe-get-one meta 'tables table 'fields name 'type))]
              (get ftype rest))))
      
      (get table fields)))
  
  (define (format-field field params)
    (let* [(name (car field))
           (tables (cadr field))
           (alias (fold-disp tables))
           (prefix (fold-disp (cdr tables)))
           (table (car (reverse tables)))
           (type (field-type field))
           ]
      ;(display table) (display " ") (display name) (display " ") (display type) (newline)
      (if type
          (if (desc? name)
              (format ",~n  ~a \"f~a~a\"" (desc-str name) prefix name)
              (format ",~n  r~a.\"f~a\" \"f~a~a\"" alias name prefix name))
          (let* [(ext-type (safe-get-one params 'ext name 'type))
                 (ext-type-ddl (type-ddl ext-type))]
            (format fmt-select-ext (format "r~a.\"fObject\"" alias) name ext-type-ddl (format "~a~a" prefix name))))))
  
  (define (format-field-inf field params)
    (let* [(name (car field))
           (tables (cadr field))
           (alias (fold-disp tables))
           (prefix (fold-disp (cdr tables)))
           (table (car (reverse tables)))
           (type (field-type field))
           ]
      ;(display table) (display " ") (display name) (display " ") (display type) (newline)
      (if type
          (if (desc? name)
              (format "~a" (desc-str name))
              (format "r~a.\"f~a\"" alias name))
          (let* [(ext-type (safe-get-one params 'ext name 'type))
                 (ext-type-ddl (type-ddl ext-type))]
            (format fmt-field-inf (format "r~a.\"fObject\"" alias) name ext-type-ddl)))))
  
  (define (format-field-def field params)
    (let* [(name (car field))
           (tables (cadr field))
           (alias (fold-disp tables))
           (table (car (reverse tables)))
           (type (field-type field))]
      (if type
          (if (desc? name)
              (format "~a" (desc-str name))
              (format "r~a.\"f~a\"" alias name))
          (let* [(ext-type (safe-get-one params 'ext name 'type))
                 (ext-type-ddl (type-ddl ext-type))]
            (format fmt-field-def-ext (format "r~a.\"fObject\"" alias) name ext-type-ddl)))))
  
  (define (format-ext-head field params)
    (let* [(name (car field))
           (tables (cadr field))
           (alias (fold-disp tables))
           (prefix (fold-disp (cdr tables)))
           (table (car (reverse tables)))
           (type (field-type field))]
      (if type
          ""
          (let* [(ext-type (safe-get-one params 'ext name 'type))
                 (ext-type-ddl (type-ddl ext-type))]
            (format "~a:~a," (format "~a~a" prefix name) ext-type-ddl)))))
  
  (define (join-table-name join)
    (define (get-table table join)
      (if (null? join)
          table
          (let* [(field (car join))
                 (over (safe-get-one meta 'tables table 'fields field 'type))]
            (unless over (error (format "field ~a is not refence" field)))
            (get-table over (cdr join))
            )))
    (get-table (car join) (cdr join)))
  
  (define (format-join join)
    (let [(over (fold-disp join))
          (base (fold-disp (reverse (cdr (reverse join)))))
          (ref (car (reverse join)))
          (table (join-table-name join))]
      (format "LEFT OUTER JOIN~n  \"t~a\" r~a ON (r~a.\"fObject\" = r~a.\"f~a\")~n" table over over base ref)))
  
  (define (format-field-ref join)
    (let [(ref (fold-disp (cdr join)))
          (table (join-table-name join))]
      (format "~a ~a " table ref)))
  
  (define (query-field-defs fields)
    (let [(params '(params))]
      (map (lambda (field) (format-field-def field params)) (fields-hier fields))))
  
  (define (query-select fields)
    (let* [(params '(params))
           (table-n (format "~a" (car fields)))
           (first (safe-get-one params 'first))
           (conditions (safe-get-one params 'conditions))
           (order (safe-get-one params 'order))
           (first-s (if first (format " FIRST ~a" first) ""))
           (where-s (if conditions
                        (string-append
                         "WHERE" br
                         "  (" (car conditions) ")"
                         (fold-strings (map format-cond (cdr conditions))) br)
                        ""))
           (order-s (if order
                        (string-append
                         "ORDER BY" br
                         "  " (car order)
                         (fold-strings (map format-ord (cdr order))))
                        ""))
           (fields-h (fields-hier fields))
           (fields-s (fold-disp (map (lambda (field) (format-field field params)) fields-h)))
           (ext-head (fold-disp (map (lambda (field) (format-ext-head field params)) fields-h)))]
      (string-append ;"--$EXT=" ext-head br
       "SELECT" first-s br
       "  r" table-n ".\"fObject\" \"fObject\""
       fields-s "/*EXT*/" br
       "FROM" br
       "  \"t" table-n "\" r" table-n br
       (fold-disp (map format-join (join-hier fields)))
       where-s
       order-s
       )))
  
  (define (query-refresh fields)
    (let* [(params '(params))
           (table-n (format "~a" (car fields)))
           (fields-s (fold-disp (map (lambda (field) (format-field field params)) (fields-hier fields))))
           (ext-head (fold-disp (map (lambda (field) (format-ext-head field params)) (fields-hier fields))))]
      (string-append ;"--$EXT=" ext-head br
       "SELECT" br
       "  r" table-n ".\"fObject\" \"fObject\""
       fields-s "/*EXT*/," br
       "  seq.\"oIdx\"" br
       "FROM" br
       "  \"pObjSeq\"(:aMode, :aObject, :aIntP0) seq" br
       "LEFT OUTER JOIN" br
       (format "  \"t~a\" r~a ON (r~a.\"fObject\" = seq.\"oObj\")" table-n table-n table-n) br
       (fold-disp (map format-join (join-hier fields)))
       )))
  
  (define (query-update fields)
    (let* [(base (base-fields fields))
           (table-n (format "~a" (car fields)))]
      (string-append "UPDATE" br
                     "  \"t" table-n "\"" br
                     "SET" br
                     (format "  \"f~a\" = :a~a" (car base) (car base))
                     (fold-disp (map (lambda (field) (format ",~n  \"f~a\" = :a~a" field field)) (cdr base))) br
                     "WHERE" br
                     "  \"fObject\" = :aObject"
                     )))
  
  (define (query-insert fields)
    (let* [(base (base-fields fields))
           (table-n (format "~a" (car fields)))]
      (string-append "INSERT INTO" br
                     "  \"t" table-n "\"" br
                     "(" br
                     "  \"fObject\"" (fold-disp (map (lambda (field) (format ", \"f~a\"" field)) base)) br
                     ") values (" br
                     "  :aObject" (fold-disp (map (lambda (field) (format ", :a~a" field)) base)) br
                     ")"
                     )))
  
  (define (query-delete fields)
    (let* [(table-n (format "~a" (car fields)))]
      (format "EXECUTE PROCEDURE \"p~aDelete\"(:aObject)" table-n)))
  
  (define (fields-inf fields)
    (let* [(params '(params))
           (table-n (format "~a" (car fields)))
           (fields-h (fields-hier fields))
           (fields-s (map (lambda (field) (format-field-inf field params)) fields-h))
           ]
      (append
       (list (string-append "r" table-n ".\"fObject\""))
       fields-s
       )))
  
  (define (query-ref fields)
    (let [(refs (fold-disp (map format-field-ref (join-hier fields))))
          (table-n (format "~a" (car fields)))]
      (string-append table-n " Object " refs)))
  
  (define (query-def fields)
    (let* [(table (car fields))
           (def (get-fields table (field-has-tag 'def) field-name))]
      (if (null? def)
          ;'Object
          (cadr fields); first field
          (car def))))
  
  (define (query-owner fields)
    (let* [(table (car fields))
           (owner (get-fields table (field-has-tag 'owner) field-name))]
      (if (null? owner)
          #f
          (car owner))))
  
  (define (get-fields table cond conv)
    (map (lambda (field) (conv field table))
         (filter (lambda (field) (cond field table))
                 (safe-get-val meta 'tables table 'fields))))
  
  (define (field-any? field table)
    #t)
  
  (define (field-has-caption? field table)
    (safe-get-val field 'caption))
  
  (define (field-has-tag-t? tag field table)
    (let [(tags (safe-get-val field 'tags))]
      (if tags
          (member tag tags)
          #f)))
  
  (define (field-has-tag tag)
    (lambda (field table)
      (field-has-tag-t? tag field table)))
  
  (define (field-in-select? field table)
    #t)
  ;(not (field-has-tag-t? 'hide-select field table)))
  
  (define (field-name field table)
    (car field))
  
  (define (field-anch field table)
    (let* [(name (car field))
           (type (safe-get-one field 'type))
           (is-ref (and (not (field-has-tag-t? 'owner field table)) (safe-get-val meta 'tables type)))]
      (if is-ref
          (cons name (get-fields type field-in-select? field-anch))
          name)))
  
  (define (table-default-fields table)
    (cons table (get-fields table field-in-select? field-anch)))
  
  (define (base-fields fields)
    (cdr (map (lambda (field) (if (list? field) (car field) field)) fields)))
  
  (define (table-base-fields table)
    (cons table (get-fields table field-in-select? field-name)))
  
  (define (table-visual-fields table)
    (get-fields table field-has-caption? field-name))
  
  (define (type-is-ref? type)
    (let [(stype (if (list? type) (car type) type))]
      (if (safe-get-one meta 'tables stype)
          #t
          #f)))
  
  (define (get-childs-of-table table-n)
    (append (filter (lambda (table)
                      (let [(own (filter (lambda (field)
                                           (and (member 'owner (safe-list (al.val field 'tags)))
                                                (eq? (al.val1 field 'type) table-n)))
                                         (al.val table 'fields)))]
                        (not (null? own))))
                    (al.val meta 'tables))
            (if (or (eq? table-n 'Extend) (eq? table-n 'Attach) (eq? table-n 'MutualRef))
                null
                (list (al.node meta 'tables 'Extend)
                      (al.node meta 'tables 'Attach)
                      (al.node meta 'tables 'MutualRef)
                      ))))
  
  (define (store-refs table n)
    (let* ([table-name (car table)]
           [fields (al.val table 'fields)]
           [refs (filter (lambda (field)
                           (let ([type (al.val1 field 'type)]
                                 [tags (al.val field 'tags)])
                             (if (al.val meta 'tables type)
                                 (if tags
                                     (not (member 'owner tags))
                                     #t)
                                 #f)))
                         fields)])
      
      (if (null? refs)
          ""
          (mix
           n "  FOR SELECT" br
           n "    \"oObj\"" br
           n "  FROM" br
           n "    \"pObjSeq\"(" br
           n "      0," br
           n "      (SELECT" br
           n "       "
           (apply mix (map (lambda (field) 
                             (let ([name (car field)])
                               (mix "t.\"f" name "\" || ")))
                           refs))
           "''" br
           n "       FROM \"t" table-name "\" t WHERE t.\"fObject\" = :lNewObject)," br
           n "      0)" br
           n "  INTO" br
           n "    :lRef" br
           n "  DO" br
           n "  BEGIN" br
           n "    IF (NOT EXISTS (SELECT 1 FROM TRANS WHERE TRNS = :lTrans AND OLDR = :lRef)) THEN" br
           n "      INSERT INTO TRANS(TRNS, OLDR, NEWR)" br
           n "      VALUES (:lTrans, :lRef, :lRef);" br
           n "  END" br
           br
           ))))
  
  (define (restore-refs table n level)
    (let* ([table-name (car table)]
           [fields (al.val table 'fields)]
           [refs (filter (lambda (field)
                           (let ([type (al.val1 field 'type)]
                                 [tags (al.val field 'tags)])
                             (if (al.val meta 'tables type)
                                 (if tags
                                     (not (member 'owner tags))
                                     #t)
                                 #f)))
                         fields)])
      
      (if (null? refs)
          ""
          (mix
           n "  UPDATE" br
           n "    \"t" table-name "\" t" br
           n "  SET" br
           (concat-str (map (lambda (field) 
                              (let ([name (car field)])
                                (mix n
                                     "    t.\"f"
                                     name
                                     "\" = COALESCE((SELECT NEWR FROM TRANS WHERE TRNS = :lTrans AND OLDR = t.\"f" name "\"), t.\"f" name "\")")))
                            refs) (mix ", " br)) br
                                                 n "  WHERE" br
                                                 n "    t.\"fObject\" = :lNewOwner" (+ 1 level) ";" br
                                                 br))))
  
  (define (copy-mutual-ref-1 table n level)
    (if (eq? level 0)
        (mix
         n "  /* MutualRef-1 */" br
         n "  FOR SELECT" br
         n "    \"fObject\"," br
         n "    \"fValue\"" br
         n "  FROM" br
         n "    \"tMutualRef\"" br
         n "  WHERE" br
         n "    \"fOwner1\" = :lOwner" (+ level 1) br
         n "  INTO" br
         n "    :lObject," br
         n "    :lRef" br
         n "  DO" br
         n "  BEGIN" br
         n "    SELECT \"oVal\" FROM ENCODE64(GEN_ID(\"gObject\", 1)) INTO :lNewObject;" br
         br
         n "    SELECT NEWR FROM TRANS WHERE TRNS = :lTrans AND OLDR = :lRef INTO :lRef;" br
         br
         n "    INSERT INTO \"tMutualRef\"" br
         n "      (\"fObject\", \"fOwner0\", \"fOwner1\", \"fName\", \"fValue\")" br
         n "    SELECT" br
         n "      :lNewObject, \"fOwner0\", :lNewOwner" (+ level 1) ", \"fName\", :lRef" br
         n "    FROM" br
         n "      \"tMutualRef\"" br
         n "    WHERE" br
         n "      \"fObject\" = :lObject;" br
         br
         n "    IF (\"aAge\" != -1) THEN" br
         n "    BEGIN" br
         n "      IF (\"aAge\" != 0) THEN" br
         n "        EXECUTE PROCEDURE \"pMutualRefTouch\"(:lNewObject, :\"aAge\", 'I');" br
         n "      ELSE" br
         n "        EXECUTE PROCEDURE \"pMutualRefTouch\"(:lNewObject, GEN_ID(\"gAge\", 1), 'I');" br
         n "    END" br
         n "  END" br
         br
         )
        ""))
  
  (define (copy-records-ddl level)
    (lambda (table)
      (let* ([n (apply mix (list-of (+ level 1) "  "))]
             [table-n (car table)]
             [childs (get-childs-of-table table-n)]
             [fields (al.val table 'fields)]
             [own (filter (lambda (field) (or (member 'owner (safe-list (al.val field 'tags))) (member 'ext-owner (safe-list (al.val field 'tags))))) fields)]
             [owner-field-n (if (eq? level 0) 'Object (if (null? own) 'Owner (caar own)))])
        
        (define (fmt-i field) (format ", \"f~a\"" (car field)))
        (define (fmt-s field) (if (eq? (car field) owner-field-n) (mix ", :lNewOwner" level) (format ", \"f~a\"" (car field))))
        
        (mix n "/* " table-n " */" br
             n "IF (\"aSuppress\" NOT LIKE '%|" table-n "|%') THEN" br
             n "BEGIN" br
             n "  FOR SELECT" br
             n "    \"fObject\"" br
             n "  FROM" br
             n "    \"t" table-n "\"" br
             n "  WHERE" br
             n "    \"f" owner-field-n "\" = :lOwner" level br
             n "  INTO" br
             n "    :lObject" br
             n "  DO" br
             n "  BEGIN" br
             n "    SELECT \"oVal\" FROM ENCODE64(GEN_ID(\"gObject\", 1)) INTO :lNewObject;" br
             (if (eq? level 0) (mix n "  \"oNew\" = lNewObject;" br) "")
             br
             n "    INSERT INTO \"t" table-n "\"" br
             n "      (\"fObject\"" (apply mix (map fmt-i fields)) ")" br
             n "    SELECT" br
             n "      :lNewObject" (apply mix (map fmt-s fields)) br
             n "    FROM" br
             n "      \"t" table-n "\"" br
             n "    WHERE" br
             n "      \"fObject\" = :lObject;" br
             br
             ;n "    UPDATE TRANS SET NEWR = :lNewObject WHERE TRNS = :lTrans AND OLDR = :lObject;" br
             n "    INSERT INTO TRANS(TRNS, OLDR, NEWR)" br
             n "    VALUES (:lTrans, :lObject, :lNewObject);" br
             br
             ;(if (> (length childs) 3) (store-refs table n) "") ;// skip Extend, Attach and MutualRef childs
             n "    IF (\"aAge\" != -1) THEN" br
             n "    BEGIN" br
             n "      IF (\"aAge\" != 0) THEN" br
             n "        EXECUTE PROCEDURE \"p" table-n "Touch\"(:lNewObject, :\"aAge\", 'I');" br
             n "      ELSE" br
             n "        EXECUTE PROCEDURE \"p" table-n "Touch\"(:lNewObject, GEN_ID(\"gAge\", 1), 'I');" br
             n "    END" br
             ;n "  \"oNew\" = lNewObject;" br
             ;n "  SUSPEND;" br
             (if (not (null? childs))
                 (mix n br
                      n "    /* CHILDS */" br
                      n "    lOwner" (+ level 1) " = lObject;" br
                      n "    lNewOwner" (+ level 1) " = lNewObject;" br
                      br
                      (apply mix (map (copy-records-ddl (+ level 1)) (get-childs-of-table table-n))) br)
                 "")
             ;"/* === " level " === */" br
             (copy-mutual-ref-1 table n level)
             (if (> (length childs) 3) (restore-refs table n level) "")
             n "  END" br
             n "END" br
             ))))
  
  (define (shallow-copy-record-proc-ddl table-n)
    (let ([table (al.node meta 'tables table-n)])
      (list t-proc (mix "p" table-n "ShallowCopy")
            (mix "ALTER PROCEDURE \"p" table-n "ShallowCopy\" (" br
                 "  \"aObject\" CHAR(12), \"aAge\" INTEGER, \"aSuppress\" VARCHAR(250))" br
                 "RETURNS" br
                 "  (\"oNew\" CHAR(12))" br
                 "AS" br
                 "  DECLARE VARIABLE lObject CHAR(12);" br
                 "  DECLARE VARIABLE lNewObject CHAR(12);" br
                 "  DECLARE VARIABLE lOwner0 CHAR(12);" br
                 "  DECLARE VARIABLE lOwner1 CHAR(12);" br
                 "  DECLARE VARIABLE lNewOwner1 CHAR(12);" br
                 "  DECLARE VARIABLE lOwner2 CHAR(12);" br
                 "  DECLARE VARIABLE lNewOwner2 CHAR(12);" br
                 "  DECLARE VARIABLE lOwner3 CHAR(12);" br
                 "  DECLARE VARIABLE lNewOwner3 CHAR(12);" br
                 "  DECLARE VARIABLE lOwner4 CHAR(12);" br
                 "  DECLARE VARIABLE lNewOwner4 CHAR(12);" br
                 br
                 "  DECLARE VARIABLE lTrans BIGINT;" br
                 "  DECLARE VARIABLE lRef CHAR(12);" br
                 "BEGIN" br
                 "  lTrans = GEN_ID(\"gObject\", 1);" br
                 br
                 "  lOwner0 = \"aObject\";" br
                 "  \"oNew\" = NULL;" br
                 br
                 ((copy-records-ddl 0) table)
                 br
                 "  DELETE FROM" br
                 "    TRANS" br
                 "  WHERE" br
                 "    TRNS = :lTrans;" br
                 br
                 "  IF (\"oNew\" IS NULL) THEN" br
                 "    EXCEPTION ERR 'Record copy error';" br
                 "  SUSPEND;" br
                 "END;"
                 ))))
  
  )
