(module sql-procedures3
  mzscheme
  
  (require "utils.scm")
  
  (provide procedures-ddl
           procedures)
  
  (define proc-name-rx (regexp "[\u0000-\u0020]*ALTER[\u0000-\u0020]+PROCEDURE[\u0000-\u0020]+\"?([0-9A-Za-z]+)\"?"))
  
  (define (proc-name ddl)
    (cadr (regexp-match proc-name-rx ddl)))
  
  (define t-proc 'procedure)
  
  (define (mk-proc ddl)
    (list t-proc (proc-name ddl) ddl))
  
  (define procedures-ddl
    (list (mix
           "ALTER PROCEDURE \"pObjSeq\" (" br
           "    \"aMode\" INTEGER, " br
           "    \"aStrP0\" VARCHAR(30000)," br
           "    \"aIntP0\" BIGINT)" br
           "RETURNS (" br
           "    \"oObj\" CHAR(12)," br
           "    \"oIdx\" INTEGER)" br
           "AS" br
           "DECLARE VARIABLE I INTEGER;" br
           "DECLARE VARIABLE C INTEGER;" br
           "BEGIN" br
           "  IF (\"aMode\" = 0) THEN" br
           "  BEGIN -- convert aSrtP0 to resultset" br
           "    i = 0;" br
           "    c = strlen(\"aStrP0\") / 12;" br
           "    WHILE (i < c) DO" br
           "    BEGIN" br
           "      \"oObj\" = substr(\"aStrP0\", 1 + i * 12, 12 + i * 12);" br
           "      \"oIdx\" = i;" br
           "      i = i + 1;" br
           "      SUSPEND;" br
           "    END" br
           "  END ELSE IF (\"aMode\" = 1) THEN" br
           "  BEGIN -- objects created later than aIntP0" br
           "    \"oIdx\" = -1;" br
           "    FOR SELECT" br
           "      \"UID\"" br
           "    FROM" br
           "      \"tObject\" obj" br
           "    WHERE" br
           "      (obj.\"fTable\" = :\"aStrP0\") and" br
           "      (obj.\"fCreated\" > :\"aIntP0\")" br
           "    INTO" br
           "      :\"oObj\"" br
           "    DO" br
           "      SUSPEND;" br
           "  END" br
           "END;"
           )
          
          (mix
           "ALTER PROCEDURE \"pPack\" (" br
           "    \"aVal\" VARCHAR(32760))" br
           "RETURNS (" br
           "    \"oPack\" VARCHAR(32760))" br
           "AS" br
           "  DECLARE VARIABLE LN INTEGER;" br
           "  DECLARE VARIABLE SZ CHAR(5);" br
           "BEGIN" br
           "  IF (\"aVal\" IS NULL) THEN" br
           "    \"oPack\" = '-';" br
           "  ELSE" br
           "  BEGIN" br
           "    \"oPack\" = \"aVal\";" br
           "    IF (strlen(\"oPack\") > 0) THEN" br
           "      \"oPack\" = rtrim(\"oPack\");" br
           "    ln = strlen(\"oPack\");" br
           "    sz = substr(ln + 100000, 2, 6);" br
           "    \"oPack\" = '|' || sz || '=' || \"oPack\";" br
           "  END" br
           "END;")
          
          (mix
           "ALTER PROCEDURE \"pLazyCopyDelete\" (" br
           "    \"aUID\" INTEGER)" br
           "AS" br
           "BEGIN" br
           "  IF (\"aUID\" > 0) THEN" br
           "  BEGIN" br
           "    DELETE FROM" br
           "      \"tLazyCopy\"" br
           "    WHERE" br
           "      \"fOwner\" = :\"aUID\";" br
           br
           "    UPDATE" br
           "      \"tObject\"" br
           "    SET" br
           "      \"fLazy\" = 0" br
           "    WHERE" br
           "      \"fLazy\" = :\"aUID\";" br
           "  END" br
           "END;")
          
          (mix
           "ALTER PROCEDURE \"pLazyCopyCreate\" (" br
           "  \"aObject\" CHAR(12))" br
           "RETURNS (" br
           "  \"oUID\" INTEGER)" br
           "AS" br
           "BEGIN" br
           "  SELECT" br
           "    \"fLazy\"" br
           "  FROM" br
           "    \"tObject\"" br
           "  WHERE" br
           "    \"UID\" = :\"aObject\"" br
           "  INTO" br
           "    :\"oUID\";" br
           br
           "  IF (\"oUID\" != 0) THEN" br
           "    EXECUTE PROCEDURE \"pLazyCopyDelete\"(:\"oUID\");" br
           br
           "  \"oUID\" = gen_id(\"gLazy\", 1);" br
           "  UPDATE" br
           "    \"tObject\"" br
           "  SET" br
           "    \"fLazy\" = :\"oUID\"" br
           "  WHERE" br
           "    \"UID\" = :\"aObject\";" br
           br
           "  SUSPEND;" br
           "END;")
          
          (mix
           "ALTER PROCEDURE \"pLazyCopyRemove\" (" br
           "  \"aObject\" CHAR(12))" br
           "RETURNS (" br
           "  \"oUID\" INTEGER)" br
           "AS" br
           "BEGIN" br
           "  SELECT" br
           "    \"fLazy\"" br
           "  FROM" br
           "    \"tObject\"" br
           "  WHERE" br
           "    \"UID\" = :\"aObject\"" br
           "  INTO" br
           "    :\"oUID\";" br
           br
           "  IF (\"oUID\" != 0) THEN" br
           "    EXECUTE PROCEDURE \"pLazyCopyDelete\"(:\"oUID\");" br
           br
           "  SUSPEND;" br
           "END;")
          
          (mix
           "ALTER PROCEDURE \"pDocGetPaymentState\" (" br
           "    \"aDoc\" CHAR(12))" br
           "RETURNS (" br
           "    \"oState\" varchar(45))" br
           "AS" br
           "declare variable lsum double precision;" br
           "declare variable lpaysum double precision;" br
           "begin" br
           "  SELECT \"fSum\" FROM \"tDoc\" WHERE \"fObject\" = :\"aDoc\" INTO :lSum;" br
           "  lPaySum = NULL;" br
           "  SELECT SUM(\"fSum\") FROM \"tPayment\" WHERE \"fDoc\" = :\"aDoc\" INTO :lPaySum;" br
           "  IF (lPaySum IS NULL) THEN" br
           "    \"oState\" = 'Payment.Empty';" br
           "  ELSE IF (lPaySum >= lSum) THEN" br
           "    \"oState\" = 'Payment.Full';" br
           "  ELSE" br
           "    \"oState\" = 'Payment.Partial';" br
           "  SUSPEND;" br
           "END;")
          
          (mix
           "ALTER PROCEDURE \"pEmptyStr\" (" br
           "    \"iValue\" INTEGER)" br
           "RETURNS (" br
           "    \"oRes\" INTEGER)" br
           "AS" br
           "begin" br
           "  \"oRes\" = 0;" br
           "  while (\"iValue\" > 0) do begin" br
           "    \"oRes\" = \"oRes\" + 1;" br
           "    \"iValue\" = \"iValue\" - 1;" br
           "    suspend;" br
           "  end" br
           "end;")
          
          (mix
           "ALTER PROCEDURE \"pMutualDefaultSet\" (" br
           "    \"aOwner1\" CHAR(12)," br
           "    \"aName\" VARCHAR(250)," br
           "    \"aValue\" CHAR(12))" br
           "AS" br
           "  DECLARE VARIABLE lObj CHAR(12);" br
           "BEGIN" br
           "  lObj = NULL;" br
           "  SELECT" br
           "    \"fObject\"" br
           "  FROM" br
           "    \"tMutualRef\"" br
           "  WHERE" br
           "    (\"fOwner0\" = '070000000001') AND (\"fOwner1\" = :\"aOwner1\") AND (\"fName\" = :\"aName\")" br
           "  INTO" br
           "    :lObj;" br
           "  IF (lObj IS NULL) THEN" br
           "  BEGIN" br
           "    SELECT \"oVal\" FROM ENCODE64(GEN_ID(\"gObject\", 1)) INTO :lObj;" br
           "    INSERT INTO \"tMutualRef\"(\"fObject\", \"fOwner0\", \"fOwner1\", \"fName\", \"fValue\")" br
           "           VALUES (:lObj, '070000000001', :\"aOwner1\", :\"aName\", :\"aValue\");" br
           "  END" br
           "  ELSE" br
           "    UPDATE" br
           "      \"tMutualRef\"" br
           "    SET" br
           "      \"fValue\" = :\"aValue\"" br
           "    WHERE" br
           "      \"fObject\" = :lObj;" br
           "END;")
          
          (mix
           "ALTER PROCEDURE \"pMutualDefaultFixate\" (" br
           "    \"aOwner0\" CHAR(12))" br
           "AS" br
           "  DECLARE VARIABLE lObj CHAR(12);" br
           "  DECLARE VARIABLE lOwner1 CHAR(12);" br
           "  DECLARE VARIABLE lValue CHAR(12);" br
           "  DECLARE VARIABLE lName VARCHAR(250);" br
           "  DECLARE VARIABLE lType VARCHAR(30);" br
           "BEGIN" br
           "  lType = NULL;" br
           "  SELECT \"fType\" FROM \"tDoc\" WHERE \"fObject\" = :\"aOwner0\" INTO :lType;" br
           "  IF (lType IS NOT NULL) THEN" br
           "  BEGIN" br
           "    FOR SELECT" br
           "      d.\"fOwner1\"," br
           "      d.\"fName\"," br
           "      d.\"fValue\"" br
           "    FROM" br
           "      \"tMutualRef\" d" br
           "    WHERE" br
           "      (d.\"fOwner0\" = '070000000001')" br
           "      AND ((d.\"fName\" || '|') CONTAINING ('.' || :lType || '|'))" br
           "      AND NOT EXISTS (" br
           "        SELECT 1" br
           "        FROM \"tMutualRef\" b" br
           "        WHERE (b.\"fOwner0\" = :\"aOwner0\") AND (b.\"fOwner1\" = d.\"fOwner1\") AND (b.\"fName\" = d.\"fName\"))" br
           "    INTO" br
           "      :lOwner1," br
           "      :lName," br
           "      :lValue" br
           "    DO" br
           "    BEGIN" br
           "      SELECT \"oVal\" FROM ENCODE64(GEN_ID(\"gObject\", 1)) INTO :lObj;" br
           "      INSERT INTO \"tMutualRef\"(\"fObject\", \"fOwner0\", \"fOwner1\", \"fName\", \"fValue\")" br
           "             VALUES (:lObj, :\"aOwner0\", :lOwner1, :lName, :lValue);" br
           "      EXECUTE PROCEDURE \"pMutualRefTouch\"(lObj, GEN_ID(\"gAge\", 1), 'I');" br
           "    END" br
           "    FOR SELECT" br
           "      m.\"fObject\"" br
           "    FROM" br
           "      \"tMutualRef\" m" br
           "    WHERE" br
           "      m.\"fOwner0\" = :\"aOwner0\"" br
           "      AND ((m.\"fName\" || '|') NOT CONTAINING ('.' || :lType || '|'))" br
           "    INTO" br
           "      :lObj" br
           "    DO" br
           "      EXECUTE PROCEDURE \"pMutualRefDelete2\"(:lObj);" br
           "  END" br
           "END;")
          
          (mix
           "ALTER PROCEDURE \"pMutualDefaultAgg\" (" br
           " N INTEGER)" br
           "AS" br
           "DECLARE VARIABLE lObj CHAR(12);" br
           "DECLARE VARIABLE lOwner1 CHAR(12);" br
           "DECLARE VARIABLE lValue CHAR(12);" br
           "DECLARE VARIABLE lName VARCHAR(250);" br
           "BEGIN" br
           "  FOR SELECT" br
           "    m.\"fOwner1\"," br
           "    m.\"fName\"" br
           "  FROM" br
           "    \"tMutualRef\" m" br
           "  WHERE" br
           "    m.\"fOwner0\" <> '070000000001'" br
           "  GROUP BY" br
           "    m.\"fOwner1\"," br
           "    m.\"fName\"" br
           "  HAVING" br
           "    COUNT(*) > 0" br
           "  ORDER BY" br
           "    m.\"fOwner1\"," br
           "    m.\"fName\"" br
           "  INTO" br
           "    :lOwner1," br
           "    :lName" br
           "  DO" br
           "  BEGIN" br
           "    lValue = NULL;" br
           "    SELECT FIRST 1" br
           "      m.\"fValue\"" br
           "    FROM" br
           "      \"tMutualRef\" m" br
           "    WHERE" br
           "      (m.\"fOwner1\" = :lOwner1)" br
           "      AND (m.\"fName\" = :lName)" br
           "    GROUP BY" br
           "      m.\"fValue\"" br
           "    ORDER BY" br
           "      COUNT(*) DESC" br
           "    INTO" br
           "      :lValue;" br
           "    IF (lValue IS NOT NULL) THEN" br
           "    BEGIN" br
           "      lObj = NULL;" br
           "      SELECT" br
           "        m.\"fObject\"" br
           "      FROM" br
           "        \"tMutualRef\" m" br
           "      WHERE" br
           "        (m.\"fOwner0\" = '070000000001')" br
           "        AND (m.\"fOwner1\" = :lOwner1)" br
           "        AND (m.\"fName\" = :lName)" br
           "      INTO" br
           "        :lObj;" br
           "      IF (lObj IS NULL) THEN" br
           "      BEGIN" br
           "        SELECT \"oVal\" FROM ENCODE64(GEN_ID(\"gObject\", 1)) INTO :lObj;" br
           "        INSERT INTO \"tMutualRef\"(\"fObject\", \"fOwner0\", \"fOwner1\", \"fName\", \"fValue\")" br
           "               VALUES (:lObj, '070000000001', :lOwner1, :lName, :lValue);" br
           "      END" br
           "      ELSE" br
           "        UPDATE" br
           "          \"tMutualRef\" m" br
           "        SET" br
           "          \"fValue\" = :lValue" br
           "        WHERE" br
           "          m.\"fObject\" = :lObj;" br
           "    END" br
           "  END" br
           "END;")
          
          (mix
           "ALTER PROCEDURE \"pDocResetPriceDC\" (" br
           "  \"aObject\" CHAR(12))" br
           "RETURNS" br
           "  (\"oItem\" CHAR(12))" br
           "AS" br
           "  DECLARE VARIABLE lPriceDC DOUBLE PRECISION;" br
           "BEGIN" br
           "  FOR SELECT" br
           "    g.\"fObject\"," br
           "    g.\"fPriceDC\"" br
           "  FROM" br
           "    \"tGoods\" g" br
           "  WHERE" br
           "    g.\"fDoc\" = :\"aObject\"" br
           "  INTO" br
           "    :\"oItem\"," br
           "    :lPriceDC" br
           "  DO" br
           "  BEGIN" br
           "    IF ((lPriceDC > 0.00001) OR (lPriceDC < -0.00001)) THEN" br
           "    BEGIN" br
           "      EXECUTE PROCEDURE \"pGoodsTouch\"(:\"oItem\", GEN_ID(\"gAge\", 1), 'U');" br
           "      UPDATE" br
           "        \"tGoods\" g" br
           "      SET" br
           "        g.\"fPriceDC\" = 0.0" br
           "      WHERE" br
           "        g.\"fObject\" = :\"oItem\";" br
           "      SUSPEND;" br
           "    END" br
           "  END" br
           "END;")
          
          (mix
           "ALTER PROCEDURE \"pFixPD4KBK\" (N INTEGER)" br
           "AS" br
           "  DECLARE VARIABLE lExtend CHAR(12);" br
           "  DECLARE VARIABLE lKBK CHAR(12);" br
           "  DECLARE VARIABLE lName VARCHAR(750);" br
           "BEGIN" br
           "  FOR SELECT" br
           "    e.\"fObject\"," br
           "    e.\"fV2\"" br
           "  FROM" br
           "    \"tDoc\" d" br
           "  JOIN" br
           "    \"tExtend\" e ON (e.\"fOwner\" = d.\"fObject\")" br
           "  WHERE" br
           "    (d.\"fType\" = 'doc-pd4')" br
           "    AND (e.\"fV2\" > '')" br
           "  INTO" br
           "    :lExtend," br
           "    :lName" br
           "  DO" br
           "  BEGIN" br
           "    lKBK = NULL;" br
           "    SELECT" br
           "      k.\"fObject\"" br
           "    FROM" br
           "      \"tKBK\" k" br
           "    WHERE" br
           "      (k.\"fTypeKBK\" = 'pd4') AND (k.\"fName\" = :lName)" br
           "    INTO" br
           "      :lKBK;" br
           br
           "    IF (lKBK IS NULL) THEN" br
           "    BEGIN" br
           "      SELECT" br
           "        \"oVal\"" br
           "      FROM" br
           "        ENCODE64(GEN_ID(\"gObject\", 1))" br
           "      INTO" br
           "        :lKBK;" br
           "      INSERT INTO \"tKBK\"(\"fObject\", \"fName\", \"fTypeKBK\")" br
           "        VALUES (:lKBK, substring(:lName from 1 for 30), 'pd4');" br
           "      EXECUTE PROCEDURE \"pKBKTouch\"(:lKBK, GEN_ID(\"gAge\", 1), 'I');" br
           "    END" br
           "    UPDATE \"tExtend\" e SET e.\"fV11\" = :lKBK WHERE e.\"fObject\" = :lExtend;" br
           "  END" br
           "END;")
          
          ))
  
  (define (procedures)
    (map mk-proc procedures-ddl))
  
  )

;(require sql-procedures3)