(module sql-utils3
  mzscheme
  
  (require
   "list.ss"
   (prefix pu. "pv-utils.scm")
   (prefix mu. "meta-utils.scm")
   "utils.scm"
   "meta-utils3.scm"
   (prefix meta. "meta.scm")
   (prefix al. "alists.scm")
   (prefix ms. "meta-sql.scm"))
  
  (provide table-ddl
           create-table-ddl
           
           table-proc-pack-ddl
           table-proc-unpack-ddl
           table-proc-touch-ddl
           table-proc-delete-ddl
           table-proc-delete2-ddl
           table-proc-fill-childs-ddl
           
           touch-object-proc-ddl
           delete-object-proc-ddl
           unpack-object-proc-ddl
           lazy-copy-restore-proc-ddl
           encode64-proc-ddl
           
           table-add-fk-ddl
           table-del-fk-ddl
           
           table-add-pk-ddl
           table-del-pk-ddl
           
           disp-alter-cmds
           alter-field-size
           desc-q-select
           
           ;test
           ;type-ddl
           )
  
  (define (false) #f)
  
  (define meta meta.raw_meta)
  
  (define t-proc 'procedure)
  
  (define (type-plain? type)
    (if (hash-table-get (mu.hlist-hash (mu.meta-get 'types)) (car type) false)
        #t
        #f))
  
  (define (type-reference? type)
    (if (hash-table-get (mu.hlist-hash (mu.meta-get 'tables)) (car type) false)
        #t
        #f))
  
  (define (type-kind type)
    (cond
      ((type-plain? type) 'plain)
      ((type-reference? type) 'reference)
      (else (error "unknown type"))))
  
  ;(define br (string #\return #\newline))
  ;(define br (string #\return))
  
  (define (plain-type->string-list plain)
    (if (null? plain)
        null
        (let* ([first (car plain)]
               [rest (cdr plain)]
               [next (if (null? rest)
                         null
                         (car rest))])
          (if (eq? first 'BLOB)
              (list "BLOB SUB_TYPE 0")
              (append
               (cond
                 ((symbol? first)
                  (list
                   (symbol->string first)
                   (cond
                     ((number? next) "(")
                     ((symbol? next) " ")
                     ((null? next) "")
                     (else (error "plain-type->string: unknown item")))))
                 ((number? first)
                  (list
                   (number->string first)
                   (if (number? next)
                       ", "
                       ")")))
                 (else (error "plain-type->string: unknown item")))
               (plain-type->string-list rest))))))
  
  ;(string-list->string (plain-type->string '(double precision)))
  ;##################################################################
  (define (field-type field)
    (hash-table-get (mu.hlist-hash field) 'type))
  
  (define (field-any? name field table)
    #t)
  
  (define (field-non-blob? name field table)
    (let* ([type (car (hash-table-get (mu.hlist-hash field) 'type false))]
           [ddl1 (al.val1 meta 'types type 'ddl)])
      (not (eq? ddl1 'BLOB))))
  
  (define (field-has-tag-t? tag name field table)
    (let [(tags (hash-table-get (mu.hlist-hash field) 'tags false))]
      (if tags
          (pu.list-has? tags tag)
          #f)))
  
  (define (field-has-tag tag)
    (lambda (name field table)
      (field-has-tag-t? tag name field table)))
  
  (define (field-name name field table)
    (list name))
  
  (define (field-val-type name field table)
    (list (field-type field)))
  
  (define (get-fields table cond conv)
    (letrec [
             (fields (mu.meta-get 'tables table 'fields))
             (fields-hash (mu.hlist-hash fields))
             (fields-names (map car (mu.hlist-list fields)))
             ]
      (flatten+
       (map
        (lambda (lrec) ; cadr - name, caddr - field
          (conv (cadr lrec) (caddr lrec) table))
        (filter car
                (map
                 (lambda (name) (let [(field (hash-table-get fields-hash name))]
                                  (list ; lrec
                                   (cond name field table)
                                   name
                                   field)))
                 fields-names))))))
  
  (define (type-ddl type)
    (case (type-kind type)
      ('plain (pu.string-list->string (plain-type->string-list (mu.meta-get 'types (car type) 'ddl))))
      ('reference (string-append "CHAR(12) /* " (symbol->string (car type)) " */"))
      (else (error "unknown type"))))
  
  (define (default-val ddl)
    (let ([first (car ddl)]
          [second (if (null? (cdr ddl)) 0 (cadr ddl))])
      (if (and (eq? first 'CHAR) (eq? second 12))
          "'000000000000'"
          (case first
            ((DOUBLE) "0.0")
            ((INTEGER) "0")
            ((BIGINT) "0")
            ((DATE TIMESTAMP) "'1899.12.30'")
            ((CHAR VARCHAR) "''")
            (else "''")
            ))))
  
  (define (type-ddl->string ddl)
    (string-append 
     (pu.string-list->string (plain-type->string-list ddl))
     " DEFAULT " (default-val ddl) " NOT NULL"))
  
  (define (create-type-ddl type)
    (case (type-kind type)
      ('plain (type-ddl->string (mu.meta-get 'types (car type) 'ddl)))
      ('reference (string-append "CHAR(12) DEFAULT '000000000000' NOT NULL /* " (symbol->string (car type)) " */"))
      (else (error "unknown type"))))
  
  (define (field-ddl field field-h)
    (string-append 
     "  \"f" (symbol->string field) "\" " (create-type-ddl (hash-table-get field-h 'type))))
  
  (define (table-ddl table)
    (letrec (
             (fields (mu.meta-get 'tables table 'fields))
             (fields-h (mu.hlist-hash fields))
             (fields-n (mu.hlist-names fields))
             (table-head
              (string-append 
               "CREATE TABLE \"t" (symbol->string table) "\" (" br
               "  \"fObject\" CHAR(12) NOT NULL," br))
             (field-break (string-append "," br)))
      
      (define (field-hash field)
        (mu.hlist-hash (hash-table-get fields-h field)))
      
      (string-append
       table-head
       (apply string-append
              (pu.insert-between
               (map 
                (lambda (field) 
                  (field-ddl field (field-hash field)))
                fields-n)
               field-break))
       br ")")))
  
  (define (create-table-ddl) (pu.string-list->string (pu.insert-after (map table-ddl (mu.hlist-names (mu.meta-get 'tables))) (string-append ";" br br "COMMIT;" br br))))
  
  (define (table-del-pk-ddl table)
    (let ([name (mix "PK_" (string-upcase (mix table)))])
      (mix "ALTER TABLE \"t" table "\" DROP CONSTRAINT " name ";" br)))
  
  (define (table-add-pk-ddl table)
    (let ([name (mix "PK_" (string-upcase (mix table)))])
      (mix "ALTER TABLE \"t" table "\" ADD CONSTRAINT " name " PRIMARY KEY (\"fObject\");" br)))
  
  ;(define (test) (table-del-pk-ddl 'Doc))
  
  (define (flatten+ a-list) ; convert list of lists to single list - version 2
    (apply append a-list))
  
  (define (field-ref? field)
    (al.node meta 'tables (al.val1 field 'type)))
  
  (define (ref-fields table)
    (filter field-ref? (al.val meta 'tables table 'fields)))
  
  (define (fk-add-ddl table)
    (lambda (field)
      (let* ([name (car field)]
             [NAME (string-upcase (mix name))]
             [type (al.val1 field 'type)]
             [TABLE (string-upcase (mix table))])
        (mix "ALTER TABLE \"t" table "\" ADD CONSTRAINT FK_" TABLE "_" NAME " "
             "FOREIGN KEY (\"f" name "\") REFERENCES \"t" type "\" (\"fObject\");" br))))
  
  (define (fk-del-ddl table)
    (lambda (field)
      (let* ([name (car field)]
             [NAME (string-upcase (mix name))]
             [type (al.val1 field 'type)]
             [TABLE (string-upcase (mix table))])
        (mix "ALTER TABLE \"t" table "\" DROP CONSTRAINT FK_" TABLE "_" NAME ";" br))))
  
  (define (table-add-fk-ddl table)
    (map (fk-add-ddl table) (ref-fields table)))
  
  (define (table-del-fk-ddl table)
    (map (fk-del-ddl table) (ref-fields table)))
  
  (define (table-any? name table)
    #t)
  
  (define (table-has-owner-o? owner name table)
    (let [(own-fields (get-fields name (field-has-tag 'owner) field-val-type))]
      (if (null? own-fields)
          #f
          (eq? owner (caar own-fields)))))
  
  (define (table-has-owner owner)
    (lambda (name table) (table-has-owner-o? owner name table)))
  
  (define (table-name name table)
    (list name))
  
  (define (get-tables cond conv)
    (flatten+
     (map (lambda (pair) (conv (car pair) (cdr pair)))
          (filter (lambda (pair) (cond (car pair) (cdr pair)))
                  (map (lambda (name) (cons name (mu.meta-get 'tables name)))
                       (mu.hlist-names (mu.meta-get 'tables)))))))
  
  (define (clear-refs table)
    (let* ([fields (al.val meta 'tables table 'fields)]
           [refs (filter field-ref? fields)]
           [ref-names (map car refs)])
      (if (null? ref-names)
          "  /* no references */"
          (mix "  UPDATE" br
               "    \"t" table "\" t" br
               "  SET" br
               (apply mix (cdr (apply append (map (lambda (name) (list ",\n" (format "    t.\"f~a\" = '000000000000'" name))) ref-names)))) br
               "  WHERE" br
               "    t.\"fObject\"  = :\"aObject\";" br
               ))))
  
  (define (table-proc-delete-ddl name)
    (let* ([t (symbol->string name)]
           [childs (get-tables (table-has-owner name) table-name)]
           [tags (al.val meta 'tables name 'tags)]
           [extend (if tags (member 'extend tags) #f)])
      (list t-proc (format "p~aDelete" name)
            (string-append
             "ALTER PROCEDURE \"p" t "Delete\" (" br
             "  \"aObject\" CHAR(12))" br
             "AS" br
             ;(if (null? childs)
             ;    ""
             (string-append "  DECLARE VARIABLE lChild CHAR(12);" br);)
             "BEGIN" br
             (clear-refs name) br
             (pu.string-list->string 
              (map 
               (lambda (child)
                 (let [(c (symbol->string child))]
                   (string-append
                    "  FOR SELECT" br
                    "    t.\"fObject\"" br
                    "  FROM" br
                    "    \"t" c "\" t" br
                    "  WHERE" br
                    "    t.\"f" (symbol->string (car (get-fields child (field-has-tag 'owner) field-name)))
                    "\" = :\"aObject\"" br
                    "  INTO" br
                    "    :lChild" br
                    "  DO" br
                    "  BEGIN" br
                    "    EXECUTE PROCEDURE \"p" c "Delete\"(:lChild);" br
                    "  END" br
                    br
                    ))) childs))
             (if (not extend)
                 (mix
                  "  FOR SELECT" br
                  "    m.\"fObject\"" br
                  "  FROM" br
                  "    \"tMutualRef\" m" br
                  "  WHERE" br
                  "    (m.\"fOwner0\" = :\"aObject\") OR (m.\"fOwner1\" = :\"aObject\")" br
                  "  INTO" br
                  "    :lChild" br
                  "  DO" br
                  "    EXECUTE PROCEDURE \"pMutualRefDelete\"(:lChild);" br
                  br
                  "  FOR SELECT" br
                  "    m.\"fObject\"" br
                  "  FROM" br
                  "    \"tMutualRef\" m" br
                  "  WHERE" br
                  "    m.\"fValue\" = :\"aObject\"" br
                  "  INTO" br
                  "    :lChild" br
                  "  DO" br
                  "  BEGIN" br
                  "    UPDATE" br
                  "      \"tMutualRef\" m" br
                  "    SET" br
                  "      m.\"fValue\" = '000000000000'" br
                  "    WHERE" br
                  "      m.\"fObject\" = :lChild;" br
                  "  END" br
                  br
                  "  DELETE FROM" br
                  "    \"tExtend\" x" br
                  "  WHERE" br
                  "    x.\"fOwner\" = :\"aObject\";" br
                  br)
                 "")
             "  DELETE FROM" br
             "    \"t" t "\" t" br
             "  WHERE" br
             "    t.\"fObject\" = :\"aObject\";" br
             "END;" br
             br
             "COMMIT;" br
             br
             ))))
  
  (define (table-proc-delete2-ddl name)
    (let* ([t (symbol->string name)]
           [childs (get-tables (table-has-owner name) table-name)]
           [tags (al.val meta 'tables name 'tags)]
           [extend (if tags (member 'extend tags) #f)])
      (list t-proc (format "p~aDelete2" name)
            (mix "ALTER PROCEDURE \"p" t "Delete2\" (" br
                 "  \"aObject\" CHAR(12))" br
                 "AS" br
                 "  DECLARE VARIABLE lChild CHAR(12);" br
                 "BEGIN" br
                 br
                 "  EXECUTE PROCEDURE \"p" t "Touch\"(:\"aObject\", GEN_ID(\"gAge\", 1), 'D');" br
                 (clear-refs name) br
                 (apply mix 
                        (map (lambda (child)
                               (mix
                                "  FOR SELECT" br
                                "    t.\"fObject\"" br
                                "  FROM" br
                                "    \"t" child "\" t" br
                                "  WHERE" br
                                "    t.\"f" (car (get-fields child (field-has-tag 'owner) field-name))
                                "\" = :\"aObject\"" br
                                "  INTO" br
                                "    :lChild" br
                                "  DO" br
                                "  BEGIN" br
                                "    EXECUTE PROCEDURE \"p" child "Delete2\"(:lChild);" br
                                "  END" br
                                br))
                             childs))
                 
                 (if (not extend)
                     (mix
                      "  FOR SELECT" br
                      "    m.\"fObject\"" br
                      "  FROM" br
                      "    \"tMutualRef\" m" br
                      "  WHERE" br
                      "    (m.\"fOwner0\" = :\"aObject\") OR (m.\"fOwner1\" = :\"aObject\")" br
                      "  INTO" br
                      "    :lChild" br
                      "  DO" br
                      "    EXECUTE PROCEDURE \"pMutualRefDelete2\"(:lChild);" br
                      br
                      "  FOR SELECT" br
                      "    m.\"fObject\"" br
                      "  FROM" br
                      "    \"tMutualRef\" m" br
                      "  WHERE" br
                      "    m.\"fValue\" = :\"aObject\"" br
                      "  INTO" br
                      "    :lChild" br
                      "  DO" br
                      "  BEGIN" br
                      "    UPDATE" br
                      "      \"tMutualRef\" m" br
                      "    SET" br
                      "      m.\"fValue\" = '000000000000'" br
                      "    WHERE" br
                      "      m.\"fObject\" = :lChild;" br
                      "    EXECUTE PROCEDURE \"pMutualRefTouch\"(:lChild, GEN_ID(\"gAge\", 1), 'D');" br
                      "  END" br
                      br
                      "  DELETE FROM" br
                      "    \"tExtend\" x" br
                      "  WHERE" br
                      "    x.\"fOwner\" = :\"aObject\";" br
                      br)
                     "")
                 "  DELETE FROM" br
                 "    \"t" t "\" t" br
                 "  WHERE" br
                 "    t.\"fObject\" = :\"aObject\";" br
                 br
                 "  FOR SELECT" br
                 "    \"fObject\"" br
                 "  FROM" br
                 "    \"tUidMap\"" br
                 "  WHERE" br
                 "    \"fValue\" = :\"aObject\"" br
                 "  INTO" br
                 "    :lChild" br
                 "  DO" br
                 "    EXECUTE PROCEDURE \"pUidMapDelete2\"(:lChild);" br
                 br
                 "END;" br
                 br
                 "COMMIT;" br
                 br
                 ))))
  
  (define (field-format fmt . params)
    (lambda (name field table)
      (let [(param-values (map (lambda (param)
                                 (case param
                                   ((name) name)
                                   ((type) (field-type field))
                                   ((type-ddl) (type-ddl (field-type field)))
                                   (else "unknown")))
                               params))]
        (list (apply format (cons fmt param-values))))))
  
  (define (field-format-proc table)
    (lambda fmt-params
      (let [(params (cons (car fmt-params) (cddr fmt-params)))
            (delim (cadr fmt-params))]
        (pu.string-list->string
         (pu.insert-between 
          (get-fields table field-non-blob? (apply field-format params))
          (format delim))))))
  
  (define (table-proc-pack-ddl table)
    (let [(fields (field-format-proc table))]
      (list t-proc (format "p~aPack" table) (string-append
                                             (format "
ALTER PROCEDURE \"p~aPack\" (" table) "
  \"aOwner\" INTEGER,
  \"aObject\" CHAR(12),
  \"aState\" CHAR(1))
AS
  DECLARE VARIABLE p VARCHAR(32700);
  DECLARE VARIABLE f VARCHAR(32700);
  DECLARE VARIABLE pl INTEGER;
  DECLARE VARIABLE fl INTEGER;
  DECLARE VARIABLE ord INTEGER; " (fields "
  DECLARE VARIABLE l~a ~a;" "" 'name 'type-ddl) "
BEGIN
  IF (NOT EXISTS (SELECT 1 FROM \"tLazyCopy\" WHERE \"fObject\" = :\"aObject\")) THEN
  BEGIN

    IF (\"aState\" = 'I') THEN
      INSERT INTO
        \"tLazyCopy\"(\"UID\", \"fOwner\", \"fObject\", \"fTable\", \"fOrd\", \"fVal\")
      VALUES" (format "
        (gen_id(\"gLazyCopy\", 1), :\"aOwner\", :\"aObject\", '~a', 0, 'I');" table) "
    ELSE
    BEGIN

      SELECT " (fields "
        t.\"f~a\"" "," 'name) "
      FROM" (format "
        \"t~a\" t" table) "
      WHERE
        t.\"fObject\" = :\"aObject\"
      INTO" (fields "
        l~a" "," 'name) ";

      p = '';
      ord = 0;
" (fields (string-append "
      EXECUTE PROCEDURE \"pPack\"(l~a) RETURNING_VALUES (f);
      pl = strlen(p);
      fl = strlen(f);
      IF ((pl + fl) > 32760) THEN
      BEGIN
        INSERT INTO
          \"tLazyCopy\"(\"UID\", \"fOwner\", \"fObject\", \"fTable\", \"fOrd\", \"fVal\")
        VALUES" (format "
          (gen_id(\"gLazyCopy\", 1), :\"aOwner\", :\"aObject\", '~a', :ord, :p);" table) "
        ord = ord + 1;
        p = '';
      END
      p = p || f;
") "" 'name) "
      INSERT INTO
        \"tLazyCopy\"(\"UID\", \"fOwner\", \"fObject\", \"fTable\", \"fOrd\", \"fVal\")
      VALUES" (format "
        (gen_id(\"gLazyCopy\", 1), :\"aOwner\", :\"aObject\", '~a', :ord, :p);" table) "

    END
  END
END;
"))))
  
  (define (table-proc-unpack-ddl table)
    (let [(fields (field-format-proc table))]
      (list t-proc (format "p~aUnpack" table) (string-append
                                               (format "
ALTER PROCEDURE \"p~aUnpack\" (" table) "
  \"aObject\" CHAR(12))
AS
  DECLARE VARIABLE s CHAR(1);
  DECLARE VARIABLE p VARCHAR(32700);
  DECLARE VARIABLE pp INTEGER;
  DECLARE VARIABLE pl INTEGER;
  DECLARE VARIABLE fl INTEGER;
  DECLARE VARIABLE ord INTEGER;" (fields "
  DECLARE VARIABLE l~a ~a;" "" 'name 'type-ddl) "
BEGIN
  ord = -1;
  pp = 0;
  pl = 0;
  s = 'N';

  SELECT
    lc.\"fVal\"
  FROM
    \"tLazyCopy\" lc
  WHERE
    (lc.\"fObject\" = :\"aObject\") AND (lc.\"fOrd\" = 0)
  INTO
    :p;

  IF (substr(p, 1, 1) = 'I') THEN
  BEGIN
    DELETE FROM" (format "
      \"t~a\"" table) "
    WHERE
      \"fObject\" = :\"aObject\";
  END
  ELSE
  BEGIN" (fields "

    IF (pp >= pl) THEN
    BEGIN
      ord = ord + 1;
      SELECT
        lc.\"fVal\"
      FROM
        \"tLazyCopy\" lc
      WHERE
        (lc.\"fObject\" = :\"aObject\") AND (lc.\"fOrd\" = :ord)
      INTO
        :p;
      pl = strlen(p);
      pp = 0;
    END

    IF (substr(p, pp + 1, pp + 1) = '|') THEN
    BEGIN
      fl = substr(p, pp + 2, pp + 6);
      l~a = substr(p, pp + 8, pp + 7 + fl);
      pp = pp + 7 + fl;
    END
    ELSE
    BEGIN
      l~a = NULL;
      pp = pp + 1;
    END" "" 'name 'name) "
" (format "
    IF (EXISTS(SELECT 1 FROM \"t~a\" WHERE \"fObject\" = :\"aObject\")) THEN" table)"
      UPDATE
        " (format "\"t~a\" t" table) "
      SET" (fields "
        t.\"f~a\" = :l~a" "," 'name 'name) "
      WHERE
        t.\"fObject\" = :\"aObject\";
    ELSE" (format "
      INSERT INTO
        \"t~a\"" table) "(\"fObject\"" (fields ", \"f~a\"" "" 'name) ")
      VALUES
        (:\"aObject\"" (fields ", :l~a" "" 'name) ");

  END
END;
"))))
  
  (define (touch-owner table field)
    (let ([name (car field)]
          [owner-table (al.val1 field 'type)])
      (mix br br
           "  SELECT \"f" name "\" FROM \"t" table "\" WHERE \"fObject\" = :\"aObject\" INTO :owner;" br
           "  EXECUTE PROCEDURE \"p" owner-table "Touch\"(owner, GEN_ID(\"gAge\", 1), 'E');"
           )))
  
  (define (table-proc-touch-ddl table)
    (let* ([meta-table (al.node meta 'tables table)]
           [owner-field (get-owner-field meta-table)]
           [touch-owner (if (and owner-field (not (eq? (al.val1 owner-field 'type) 'Reference))) (touch-owner table owner-field) "")]
           [fields (field-format-proc table)])
      (list t-proc (format "p~aTouch" table)
            (mix "
ALTER PROCEDURE \"p" table "Touch\" (
  \"aObject\" CHAR(12),
  \"aAge\" BIGINT,
  \"aState\" CHAR(1))
AS
  DECLARE VARIABLE lazy INTEGER;
  DECLARE VARIABLE owner CHAR(12);
BEGIN
  IF (:\"aState\" = 'I') THEN
    INSERT INTO
      \"tObject\"(\"UID\", \"fAge\", \"fCreated\", \"fChanged\", \"fTable\", \"fState\")
    VALUES
      (:\"aObject\", :\"aAge\", :\"aAge\", CURRENT_TIMESTAMP, '" table "', 'E');
  ELSE
    IF (:\"aState\" = 'D') THEN
      UPDATE
        \"tObject\"
      SET
        \"fAge\" = :\"aAge\",
        \"fChanged\" = CURRENT_TIMESTAMP,
        \"fState\" = 'D'
      WHERE
        \"UID\" = :\"aObject\";
    ELSE
    BEGIN
      UPDATE
        \"tObject\"
      SET
        \"fChanged\" = CURRENT_TIMESTAMP
      WHERE
        \"UID\" = :\"aObject\";

      UPDATE
        \"tObject\"
      SET
        \"fAge\" = :\"aAge\"
      WHERE
        (\"UID\" = :\"aObject\") AND (\"fAge\" > 0);
    END

  SELECT
    \"oLazy\"
  FROM
    \"p" table "GetLazy\"(:\"aObject\")
  INTO
    :lazy;

  IF (lazy > 0) THEN
  BEGIN
    EXECUTE PROCEDURE \"p" table "Pack\"(:lazy, :\"aObject\", :\"aState\");
  END" touch-owner "
END;
"))))
  
  (define (touch-object-proc-ddl)
    (list t-proc "pObjectTouch" (string-append "
ALTER PROCEDURE \"pObjectTouch\" (
  \"aObject\" CHAR(12),
  \"aTable\" VARCHAR(30),
  \"aAge\" BIGINT,
  \"aState\" CHAR(1))
AS
BEGIN
" (pu.string-list->string (map (lambda (table) (format "
  IF (\"aTable\" = '~a') THEN
    EXECUTE PROCEDURE \"p~aTouch\"(:\"aObject\", :\"aAge\", :\"aState\");
  ELSE" table table)) (get-tables table-any? table-name))) "
    EXCEPTION ERR 'pObjectTouch: table \"' || COALESCE(\"aTable\", 'unknown') || '\" not defined';
END;
")))
  
  (define (delete-object-proc-ddl)
    (list t-proc "pObjectDelete" (string-append "
ALTER PROCEDURE \"pObjectDelete\" (
  \"aObject\" CHAR(12))
AS
  DECLARE VARIABLE lTable VARCHAR(30);
  DECLARE VARIABLE lAge BIGINT;
BEGIN
  SELECT
    \"fTable\"
  FROM
    \"tObject\"
  WHERE
    \"UID\" = :\"aObject\"
  INTO
    :lTable;
    
  lAge = GEN_ID(\"gAge\", 1);
  EXECUTE PROCEDURE \"pObjectTouch\"(:\"aObject\", :lTable, :lAge, 'D');
" (pu.string-list->string (map (lambda (table) (format "
  IF (lTable = '~a') THEN
    EXECUTE PROCEDURE \"p~aDelete\"(:\"aObject\");
  ELSE" table table)) (get-tables table-any? table-name))) "
    EXCEPTION ERR 'pObjectDelete: table \"' || COALESCE(lTable, 'unknown') || '\" not defined';
END;
")))
  
  (define (unpack-object-proc-ddl)
    (list t-proc "pObjectUnpack" (string-append "
ALTER PROCEDURE \"pObjectUnpack\" (
    \"aObject\" CHAR(12))
AS
  DECLARE VARIABLE lTable VARCHAR(30);
BEGIN
  SELECT
    \"fTable\"
  FROM
    \"tObject\"
  WHERE
    \"UID\" = :\"aObject\"
  INTO
    :lTable;
" (pu.string-list->string (map (lambda (table) (format "
  IF (lTable = '~a') THEN
    EXECUTE PROCEDURE \"p~aUnpack\"(:\"aObject\");
  ELSE" table table)) (get-tables table-any? table-name))) "
    EXCEPTION ERR 'pObjectUnpack: table \"' || COALESCE(lTable, 'unknown') || '\" not defined';
END;
")))
  
  (define (lazy-copy-restore-proc-ddl)
    (list t-proc "pLazyCopyRestore" "
ALTER PROCEDURE \"pLazyCopyRestore\" (
    \"aUID\" INTEGER)
RETURNS
  (\"oObject\" CHAR(12), \"oState\" CHAR(1))
AS
DECLARE VARIABLE lObject CHAR(12);
DECLARE VARIABLE lVal VARCHAR(32700);
BEGIN
  IF (\"aUID\" > 0) THEN
  BEGIN
    FOR SELECT
      \"fObject\",
      \"fVal\"
    FROM
      \"tLazyCopy\"
    WHERE
      \"fOwner\" = :\"aUID\"
    ORDER BY
      \"UID\" DESC
    INTO
      :lObject,
      :lVal
    DO
    BEGIN
      EXECUTE PROCEDURE \"pObjectUnpack\"(:lObject);
      IF (substr(lVal, 1, 1) = 'I') THEN
      BEGIN
        \"oState\" = 'D';
        DELETE FROM
          \"tObject\"
        WHERE
          \"UID\" = :lObject;
      END
      ELSE
      BEGIN
        \"oState\" = 'U';
        UPDATE
          \"tObject\"
        SET
          \"fAge\" = gen_id(\"gAge\", 1),
          \"fChanged\" = current_timestamp,
          \"fState\" = 'E'
        WHERE
          \"UID\" = :lObject;
        \"oObject\" = lObject;
        SUSPEND;
      END
    END

    DELETE FROM
      \"tLazyCopy\"
    WHERE
      \"fOwner\" = :\"aUID\";

    UPDATE
      \"tObject\"
    SET
      \"fLazy\" = 0
    WHERE
      \"fLazy\" = :\"aUID\";
  END
END;
"))
  
  (define (encode64-proc-ddl)
    (list t-proc "ENCODE64"
          (mix
           "ALTER PROCEDURE ENCODE64 (" br
           "  \"aVal\" BIGINT)" br
           "RETURNS (" br
           "  \"oVal\" CHAR(12))" br
           "AS" br
           "  DECLARE VARIABLE cTab CHAR(64);" br
           "  DECLARE VARIABLE v BIGINT;" br
           "  DECLARE VARIABLE b BIGINT;" br
           "  DECLARE VARIABLE d BIGINT;" br
           "  DECLARE VARIABLE m BIGINT;" br
           "  DECLARE VARIABLE r INTEGER;" br
           "  DECLARE VARIABLE e VARCHAR(12);" br
           "BEGIN" br
           "  cTab = '0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZ_abcdefghijklmnopqrstuvwxyz~';" br
           "  b = 64;" br
           "  v = \"aVal\";" br
           "  r = 12;" br
           "  e = '';" br
           "  WHILE (r > 0) DO" br
           "  BEGIN" br
           "    d = v / b;" br
           "    m = v - (d * b);" br
           "    v = d;" br
           "    e = substr(cTab, m + 1, m + 1) || e;" br
           "    r = r - 1;" br
           "  END" br
           "  \"oVal\" = e;" br
           "  SUSPEND;" br
           "END;")))
  
  (define (field-is-owner? field)
    (let ([tags (al.val field 'tags)])
      (if tags
          (member 'owner tags)
          #f)))
  
  (define (table-owned-by? table owner)
    (let ([owner-fields (filter field-is-owner? (al.val table 'fields))])
      (if (null? owner-fields)
          #f
          (eq? (caar owner-fields) owner))))
  
  (define (table-owned-by owner)
    (lambda (table)
      (table-owned-by? table owner)))
  
  (define (tables-owned-by owner)
    (filter (table-owned-by owner) (al.val meta 'tables)))
  
  (define (clear-child-content table)
    (let* ([fields (al.val table 'fields)]
           [owner-field (car (filter field-is-owner? fields))]
           [tn (car table)]
           [fn (car owner-field)])
      (mix "    FOR SELECT" br
           "      t.\"fObject\"" br
           "    FROM" br
           "      \"t" tn "\" t" br
           "    WHERE" br
           "      t.\"f" fn "\" = :\"aDst\"" br
           "    INTO" br
           "      :lChild" br
           "    DO" br
           "      EXECUTE PROCEDURE \"p" tn "Delete2\"(:lChild);" br
           )))
  
  (define (copy-child-content table)
    (let* ([fields (al.val table 'fields)]
           [owner-field (car (filter field-is-owner? fields))]
           [tn (car table)]
           [fn (car owner-field)])
      (mix "    INSERT INTO \"t" tn "\"" br
           "      (\"fObject\"" (apply mix (map (lambda (field) (mix ", \"f" (car field) "\"")) fields)) ")" br
           "    SELECT" br
           "      (SELECT \"oVal\" FROM ENCODE64(GEN_ID(\"gObject\", 1)))" br
           "      " (apply mix (map (lambda (field) (if (eq? (car field) fn) ", :\"aDst\"" (mix ", \"f" (car field) "\""))) fields)) br
           "    FROM" br
           "      \"t" tn "\" t" br
           "    WHERE" br
           "      t.\"f" fn "\" = :\"aSrc\";" br
           )))
  
  (define (touch-child-content table)
    (let* ([fields (al.val table 'fields)]
           [owner-field (car (filter field-is-owner? fields))]
           [tn (car table)]
           [fn (car owner-field)])
      (mix "    FOR SELECT" br
           "      t.\"fObject\"" br
           "    FROM" br
           "      \"t" tn "\" t" br
           "    WHERE" br
           "      t.\"f" fn "\" = :\"aDst\"" br
           "    INTO" br
           "      :lChild" br
           "    DO" br
           "      EXECUTE PROCEDURE \"p" tn "Touch\"(:lChild, GEN_ID(\"gAge\", 1), 'I');" br
           )))
  
  (define (table-proc-fill-childs-ddl table)
    (let ([child-tables (tables-owned-by table)])
      (list t-proc (mix "p" table "FillChilds")
            (mix "ALTER PROCEDURE \"p" table "FillChilds\"(\"aSrc\" CHAR(12), \"aDst\" CHAR(12), \"aSel\" VARCHAR(1000))" br
                 "AS" br
                 "  DECLARE VARIABLE lChild CHAR(12);" br
                 "BEGIN" br
                 (apply mix
                        (map (lambda (child)
                               (mix "  IF ((\"aSel\" = '') OR (\"aSel\" CONTAINING '" (car child) ";'))"
                                    ;" OR ((\"aSel\" CONTAINING 'except;') AND (\"aSel\" NOT CONTAINING '" (car child) ";'))
                                    " THEN" br
                                    "  BEGIN" br
                                    (clear-child-content child) br
                                    (copy-child-content child) br
                                    (touch-child-content child) br
                                    "  END" br
                                    br))
                             child-tables))
                 "END;" br))))
  
  (define (field-default field)
    (if (field-ref? field)
        "'000000000000'"
        (default-val (al.val meta 'types (al.val1 field 'type) 'ddl))))
  
  (define (new-fields old-table new-table)
    (let ([old-fields (al.val old-table 'fields)]
          [new-fields (al.val new-table 'fields)])
      (filter (lambda (field) (not (assq (car field) old-fields))) new-fields)))
  
  (define (add-field-ddl table)
    (lambda (field)
      (let* ([name (car field)]
             [type (al.val1 field 'type)]
             [ddl-s (ms.type-ddl type)]
             [default (field-default field)])
        (mix "ALTER TABLE \"t" table "\" ADD \"f" name "\" " ddl-s " DEFAULT " default " NOT NULL;" br))))
  
  (define (init-field-dml table)
    (lambda (field)
      (let* ([name (car field)]
             [default (field-default field)])
        (mix "UPDATE \"t" table "\" SET \"f" name "\" = " default ";" br))))
  
  (define (del-field-ddl table)
    (lambda (field)
      (let* ([name (car field)])
        (mix "ALTER TABLE \"t" table "\" DROP \"f" name "\";" br))))
  
  (define (table-add-field-ddl old-table new-table)
    (map (add-field-ddl (car new-table)) (new-fields old-table new-table)))
  
  (define (table-init-field-dml old-table new-table)
    (map (init-field-dml (car new-table)) (new-fields old-table new-table)))
  
  (define (table-del-field-ddl old-table new-table)
    (map (del-field-ddl (car new-table)) (new-fields new-table old-table)))
  
  (define (disp-alter-cmds old-meta new-meta)
    (lambda (table)
      (let* ([t-o (al.node old-meta 'tables table)]
             [t-n (al.node new-meta 'tables table)])
        (if t-o
            (let ([add (table-add-field-ddl t-o t-n)]
                  [init (table-init-field-dml t-o t-n)]
                  [del (table-del-field-ddl t-o t-n)])
              (if (not (null? add))
                  (begin (display (apply mix add))
                         (display "COMMIT;")
                         (newline)))
              (if (not (null? init))
                  (begin (display (apply mix init))
                         (display "COMMIT;")
                         (newline)))
              (if (not (null? del))
                  (begin (display (apply mix del))
                         (display "COMMIT;")
                         (newline))))
            (begin (display (table-ddl table)) (display ";")
                   (newline)
                   (display "COMMIT;")
                   (newline)
                   (display (format "INSERT INTO \"t~a\"(\"fObject\") VALUES ('000000000000');" table))
                   (newline))))))
  
  (define (alter-field-size meta-old meta-new old-table new-table)
    (apply string-append
           (filter (lambda (e) e)
                   (map (lambda (field)
                          (let ([old-field (al.node old-table 'fields (car field))])
                            (if old-field
                                (let* ([old-type (al.val1 old-field 'type)]
                                       [old-meta-type (al.val meta-old 'types old-type 'ddl)]
                                       [type (al.val1 field 'type)]
                                       [meta-type (al.val meta-new 'types type 'ddl)])
                                  (if (and old-meta-type meta-type)
                                      (if (and (eq? (car old-meta-type) 'VARCHAR) (eq? (car meta-type) 'VARCHAR))
                                          (let ([old-len (cadr old-meta-type)]
                                                [len (cadr meta-type)])
                                            (if (> len old-len)
                                                (mix "ALTER TABLE \"t" (car new-table) "\" ALTER \"f" (car field) "\" TYPE VARCHAR(" len ");" br
                                                     "COMMIT;" br)
                                                #f))
                                          #f)
                                      #f))
                                #f)))
                        (al.val new-table 'fields)))))
  
  ; dep = (table (dep-field1 dep-field2) desc-field description)
  (define (desc-q-select dep)
    (let* ([table (list-ref dep 0)]
           [dep-fields (list-ref dep 1)]
           [desc-field (list-ref dep 2)]
           [desc (list-ref dep 3)]
           [desc-type (al.val1 meta 'tables table 'fields desc-field 'type)]
           [desc-refer (al.node meta 'tables desc-type)]
           [refer-desc (if desc-refer
                           (mix "(SELECT r.\"f" (get-desc-field desc-refer) "\" FROM \"t" desc-type "\" r WHERE r.\"fObject\" = b.\"f" desc-field "\")")
                           (mix "b.\"f" desc-field "\""))]
           [refer-type (if desc-refer
                           (car desc-refer)
                           '-)]
           [cond (concat-str (map (lambda (field) (mix "(b.\"f" field "\" = :aObject)")) dep-fields) " OR ")])
      (mix "SELECT" br
           "  b.\"fObject\"," br
           "  '" refer-type "' \"fReferType\"," br
           "  b.\"f" desc-field "\" \"fDesc\"," br
           "  " refer-desc " \"fReferDesc\"" br
           "FROM" br
           "  \"t" table "\" b" br
           "WHERE" br
           "  " cond br)))
  
  )
