(module utils
  mzscheme
  
  (require "list.ss")
  
  (provide br
           w-br
           vector-pos
           safe-cdr
           safe-cadr
           concat-str
           mix
           mix1
           flatten
           includes
           safe-eval
           safe-substr
           safe-subbytes
           safe-list
           safe-assoc-val
           subr
           subl
           
           list#
           refine
           
           caption-conv
           encode-caption
           dmap
           encode-exp
           list-of
           param-val
           string->bytes1251
           bytes1251->string
           double->str2)
  
  (define br "\n")
  
  (define (w-br)
    (write-bytes #"\x0D\x0A"))
  
  (define (vector-pos v item)
    
    (let ([len  (vector-length v)])
      
      (define (v-p start)
        (if (<= len start)
            #f
            (if (eq? (vector-ref v start) item)
                start
                (v-p (+ start 1)))))
      
      (v-p 0)))
  
  (define (safe-cdr v)
    (if (and v (not (null? v)))
        (cdr v)
        #f))
  
  (define (safe-cadr v)
    (if (and v (not (null? v)) (not (null? (cdr v))))
        (cadr v)
        #f))
  
  (define (concat-str slist sep)
    (if (null? slist)
        ""
        (apply string-append (car slist) (cdr (map (lambda (s) (string-append sep s)) slist)))))
  
  (define (mix . any)
    (apply string-append (map (lambda (a) (format "~a" a)) any)))
  
  (define (mix1 alist)
    (apply string-append (map (lambda (a) (format "~a" a)) alist)))
  
  (define (flatten alist)
    
    (define (flatten0 ds pad)
      (if (null? ds)
          (if (null? pad)
              null
              (flatten0 (car pad) (cdr pad)))
          (let ([head (car ds)]
                [tail (cdr ds)])
            (if (list? head)
                (flatten0 head (cons tail pad))
                (cons head (flatten0 tail pad))))))
    
    (flatten0 alist null))
  
  (define (includes list items)
    (if (andmap (lambda (item) (member item list)) items) #t #f))
  
  (define (safe-eval symbol)
    (with-handlers ([exn:fail?
                     (lambda (exn) #f)])
      (eval symbol)))
  
  (define (safe-substr str start end)
    (let* ([s-str (if (string? str) str (format "~a" str))]
           [s-len (string-length s-str)]
           [s-start (if (> start s-len) s-len start)]
           [s-end (if (> end s-len) s-len end)])
      (substring s-str s-start s-end)))
  
  (define (safe-subbytes str start end)
    (let* ([s-str (if (bytes? str) str (error "byte string required"))]
           [s-len (bytes-length s-str)]
           [s-start (if (> start s-len) s-len start)]
           [s-end (if (> end s-len) s-len end)])
      (subbytes s-str s-start s-end)))
  
  (define (subr str start end)
    (if (bytes? str)
        (safe-subbytes str start end)
        (safe-substr str start end)))
  
  (define (subl str length)
    (if (bytes? str)
        (safe-subbytes str 0 length)
        (safe-substr str 0 length)))
  
  (define caption-conv
    (list
     (cons (regexp "\\\\") "\\\\B")
     (cons (regexp "\\/") "\\\\F")
     (cons (regexp "\\,") "\\\\C")
     (cons (regexp "\\:") "\\\\L")
     (cons (regexp "\\.") "\\\\D")
     (cons (regexp "\\;") "\\\\K")
     (cons (regexp " ") "\\\\S")
     (cons (regexp "\n") "\\\\N")
     (cons (regexp "\\(") "\\\\O")
     (cons (regexp "\\)") "\\\\P")))
  
  (define (encode str conv)
    (if (null? conv)
        str
        (let [(row (car conv))]
          (encode (regexp-replace* (car row) str (cdr row)) (cdr conv)))))
  
  (define (encode-caption str)
    (encode str caption-conv))
  
  (define (dmap proc data)
    (cond
      ((null? data)
       null)
      ((pair? data)
       (let [(first (car data))
             (rest (cdr data))]
         (cons (dmap proc first) (dmap proc rest))))
      (else
       (proc data))))
  
  (define (encode-exp exp)
    (dmap (lambda (e) (if (string? e) (encode-caption e) e)) exp))
  
  (define (list# . items)
    (filter (lambda (item) (not (or (void? item) (null? item)))) items))
  
  (define (refine tree)
    (if (list? tree)
        (filter (lambda (item) (not (or (void? item) (null? item))))
                (map refine tree))
        tree))
  
  (define (safe-list l)
    (if l l null))
  
  (define (list-of num item)
    (if (< num 1)
        null
        (cons item (list-of (- num 1) item))))
  
  (define (param-val name params)
    (let ([param (assoc name params)])
      (if param
          (cadr param)
          #f)))
  
  (define (c-utf8->cp1251 n)
    (if (<= n 127)
        n
        (if (and (>= n 1040) (<= n 1103))
            (- n 848)
            (case n
              ((1025) 168)
              ((1105) 184)
              ((8470) 185)
              (else 63)))))
  
  (define (string->bytes1251 s)
    (list->bytes (map (lambda (c) (c-utf8->cp1251 (char->integer c))) (string->list s))))
  
  (define (c-cp1251->utf8 n)
    (if (<= n 127)
        n
        (if (and (>= n 192) (<= n 255))
            (+ n 848)
            (case n
              ((168) 1025)
              ((184) 1105)
              ((185) 8470)
              (else 63)))))
  
  (define (bytes1251->string s)
    (list->string (map (lambda (c) (integer->char (c-cp1251->utf8 c))) (bytes->list s))))
  
  (define (safe-assoc-val item alist)
    (if alist
        (let ([val (assoc item alist)])
          (if val
              (cadr val)
              ""))
        ""))
  
  (define (double->str2 d)
  (let* ([exact (inexact->exact (round (* 100.0 d)))]
         [str (number->string exact)])
    (cond ((< exact 1) "0.00")
          ((< exact 10) (string-append "0.0" str))
          ((< exact 100) (string-append "0." str))
          (else 
           (let* ([len (string-length str)]
                  [frac (- len 2)])
             (string-append
              (substring str 0 frac)
              "."
              (substring str frac len)))))))
  
  )
